# -*- coding: utf-8 -*-
# srpauth.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import binascii
import threading
import sys

import requests
import srp
import json

# this error is raised from requests
from simplejson.decoder import JSONDecodeError
from functools import partial
from requests.adapters import HTTPAdapter

from twisted.internet import threads
from twisted.internet.defer import CancelledError

from leap.bitmask.backend.settings import Settings
from leap.bitmask.logs.utils import get_logger
from leap.bitmask.util import request_helpers as reqhelper
from leap.bitmask.util.compat import requests_has_max_retries
from leap.bitmask.util.constants import REQUEST_TIMEOUT
from leap.common.check import leap_assert
from leap.common.events import emit, catalog

logger = get_logger()


class SRPAuthenticationError(Exception):
    """
    Exception raised for authentication errors
    """
    pass


class SRPAuthConnectionError(SRPAuthenticationError):
    """
    Exception raised when there's a connection error
    """
    pass


class SRPAuthBadStatusCode(SRPAuthenticationError):
    """
    Exception raised when we received an unknown bad status code
    """
    pass


class SRPAuthNoSalt(SRPAuthenticationError):
    """
    Exception raised when we don't receive the salt param at a
    specific point in the auth process
    """
    pass


class SRPAuthNoB(SRPAuthenticationError):
    """
    Exception raised when we don't receive the B param at a specific
    point in the auth process
    """
    pass


class SRPAuthBadDataFromServer(SRPAuthenticationError):
    """
    Generic exception when we receive bad data from the server.
    """
    pass


class SRPAuthJSONDecodeError(SRPAuthenticationError):
    """
    Exception raised when there's a problem decoding the JSON content
    parsed as received from th e server.
    """
    pass


class SRPAuthBadUserOrPassword(SRPAuthenticationError):
    """
    Exception raised when the user provided a bad password to auth.
    """
    pass


class SRPAuthVerificationFailed(SRPAuthenticationError):
    """
    Exception raised when we can't verify the SRP data received from
    the server.
    """
    pass


class SRPAuthNoSessionId(SRPAuthenticationError):
    """
    Exception raised when we don't receive a session id from the
    server.
    """
    pass


class SRPAuthImpl(object):
    """
    Implementation of the SRPAuth interface
    """

    LOGIN_KEY = "login"
    A_KEY = "A"
    CLIENT_AUTH_KEY = "client_auth"
    SESSION_ID_KEY = "_session_id"
    USER_VERIFIER_KEY = 'user[password_verifier]'
    USER_SALT_KEY = 'user[password_salt]'
    AUTHORIZATION_KEY = "Authorization"

    def __init__(self, provider_config):
        """
        Constructor for SRPAuth implementation

        :param provider_config: ProviderConfig needed to authenticate.
        :type provider_config: ProviderConfig
        """
        leap_assert(provider_config,
                    "We need a provider config to authenticate")

        self._provider_config = provider_config
        self._settings = Settings()

        # **************************************************** #
        # Dependency injection helpers, override this for more
        # granular testing
        self._fetcher = requests
        self._srp = srp
        self._hashfun = self._srp.SHA256
        self._ng = self._srp.NG_1024
        # **************************************************** #

        self._reset_session()

        self._session_id = None
        self._session_id_lock = threading.Lock()
        self._uuid = None
        self._uuid_lock = threading.Lock()
        self._token = None
        self._token_lock = threading.Lock()

        self._srp_user = None
        self._srp_a = None

        # User credentials stored for password changing checks
        self._username = None
        self._password = None

    def _reset_session(self):
        """
        Resets the current session and sets max retries to 30.
        """
        self._session = self._fetcher.session()
        # We need to bump the default retries, otherwise logout
        # fails most of the times
        # NOTE: This is a workaround for the moment, the server
        # side seems to return correctly every time, but it fails
        # on the client end.
        if requests_has_max_retries:
            adapter = HTTPAdapter(max_retries=30)
        else:
            adapter = HTTPAdapter()
        self._session.mount('https://', adapter)

    def _safe_unhexlify(self, val):
        """
        Rounds the val to a multiple of 2 and returns the
        unhexlified value

        :param val: hexlified value
        :type val: str

        :rtype: binary hex data
        :return: unhexlified val
        """
        return binascii.unhexlify(val) \
            if (len(val) % 2 == 0) else binascii.unhexlify('0' + val)

    def _authentication_preprocessing(self, username, password):
        """
        Generates the SRP.User to get the A SRP parameter

        :param username: username to login
        :type username: str
        :param password: password for the username
        :type password: str
        """
        logger.debug("Authentication preprocessing...")

        self._srp_user = self._srp.User(username.encode('utf-8'),
                                        password.encode('utf-8'),
                                        self._hashfun, self._ng)
        _, A = self._srp_user.start_authentication()

        self._srp_a = A

    def _start_authentication(self, _, username):
        """
        Sends the first request for authentication to retrieve the
        salt and B parameter

        Might raise all SRPAuthenticationError based:
          SRPAuthenticationError
          SRPAuthConnectionError
          SRPAuthBadStatusCode
          SRPAuthNoSalt
          SRPAuthNoB

        :param _: IGNORED, output from the previous callback (None)
        :type _: IGNORED
        :param username: username to login
        :type username: str

        :return: salt and B parameters
        :rtype: tuple
        """
        logger.debug("Starting authentication process...")
        try:
            auth_data = {
                self.LOGIN_KEY: username,
                self.A_KEY: binascii.hexlify(self._srp_a)
            }
            sessions_url = "%s/%s/%s/" % \
                (self._provider_config.get_api_uri(),
                 self._provider_config.get_api_version(),
                 "sessions")

            ca_cert_path = self._provider_config.get_ca_cert_path()
            ca_cert_path = ca_cert_path.encode(sys.getfilesystemencoding())

            init_session = self._session.post(sessions_url,
                                              data=auth_data,
                                              verify=ca_cert_path,
                                              timeout=REQUEST_TIMEOUT)
            # Clean up A value, we don't need it anymore
            self._srp_a = None
        except requests.exceptions.ConnectionError as e:
            logger.error("No connection made (salt): {0!r}".format(e))
            raise SRPAuthConnectionError()
        except Exception as e:
            logger.error("Unknown error: %r" % (e,))
            raise SRPAuthenticationError()

        content, mtime = reqhelper.get_content(init_session)

        if init_session.status_code not in (200,):
            logger.error("No valid response (salt): "
                         "Status code = %r. Content: %r" %
                         (init_session.status_code, content))
            if init_session.status_code == 422:
                logger.error("Invalid username or password.")
                raise SRPAuthBadUserOrPassword()

            logger.error("There was a problem with authentication.")
            raise SRPAuthBadStatusCode()

        json_content = json.loads(content)
        salt = json_content.get("salt", None)
        B = json_content.get("B", None)

        if salt is None:
            logger.error("The server didn't send the salt parameter.")
            raise SRPAuthNoSalt()
        if B is None:
            logger.error("The server didn't send the B parameter.")
            raise SRPAuthNoB()

        return salt, B

    def _process_challenge(self, salt_B, username):
        """
        Given the salt and B processes the auth challenge and
        generates the M2 parameter

        Might raise SRPAuthenticationError based:
          SRPAuthenticationError
          SRPAuthBadDataFromServer
          SRPAuthConnectionError
          SRPAuthJSONDecodeError
          SRPAuthBadUserOrPassword

        :param salt_B: salt and B parameters for the username
        :type salt_B: tuple
        :param username: username for this session
        :type username: str

        :return: the M2 SRP parameter
        :rtype: str
        """
        logger.debug("Processing challenge...")
        try:
            salt, B = salt_B
            unhex_salt = self._safe_unhexlify(salt)
            unhex_B = self._safe_unhexlify(B)
        except (TypeError, ValueError) as e:
            logger.error("Bad data from server: %r" % (e,))
            raise SRPAuthBadDataFromServer()
        M = self._srp_user.process_challenge(unhex_salt, unhex_B)

        auth_url = "%s/%s/%s/%s" % (self._provider_config.get_api_uri(),
                                    self._provider_config.
                                    get_api_version(),
                                    "sessions",
                                    username)

        auth_data = {
            self.CLIENT_AUTH_KEY: binascii.hexlify(M)
        }

        try:
            auth_result = self._session.put(auth_url,
                                            data=auth_data,
                                            verify=self._provider_config.
                                            get_ca_cert_path(),
                                            timeout=REQUEST_TIMEOUT)
        except requests.exceptions.ConnectionError as e:
            logger.error("No connection made (HAMK): %r" % (e,))
            raise SRPAuthConnectionError()

        try:
            content, mtime = reqhelper.get_content(auth_result)
        except JSONDecodeError:
            logger.error("Bad JSON content in auth result.")
            raise SRPAuthJSONDecodeError()

        if auth_result.status_code == 422:
            error = ""
            try:
                error = json.loads(content).get("errors", "")
            except ValueError:
                logger.error("Problem parsing the received response: %s"
                             % (content,))
            except AttributeError:
                logger.error("Expecting a dict but something else was "
                             "received: %s", (content,))
            logger.error("[%s] Wrong password (HAMK): [%s]" %
                         (auth_result.status_code, error))
            raise SRPAuthBadUserOrPassword()

        if auth_result.status_code not in (200,):
            logger.error("No valid response (HAMK): "
                         "Status code = %s. Content = %r" %
                         (auth_result.status_code, content))
            raise SRPAuthBadStatusCode()

        return json.loads(content)

    def _extract_data(self, json_content):
        """
        Extracts the necessary parameters from json_content (M2,
        id, token)

        Might raise SRPAuthenticationError based:
          SRPBadDataFromServer

        :param json_content: Data received from the server
        :type json_content: dict
        """
        try:
            M2 = json_content.get("M2", None)
            uuid = json_content.get("id", None)
            token = json_content.get("token", None)
        except Exception as e:
            logger.error(e)
            raise SRPAuthBadDataFromServer()

        self.set_uuid(uuid)
        self.set_token(token)

        if M2 is None or self.get_uuid() is None:
            logger.error("Something went wrong. Content = %r" %
                         (json_content,))
            raise SRPAuthBadDataFromServer()

        emit(catalog.CLIENT_UID, uuid)  # make the rpc call async

        return M2

    def _verify_session(self, M2):
        """
        Verifies the session based on the M2 parameter. If the
        verification succeeds, it sets the session_id for this
        session

        Might raise SRPAuthenticationError based:
          SRPAuthBadDataFromServer
          SRPAuthVerificationFailed

        :param M2: M2 SRP parameter
        :type M2: str
        """
        logger.debug("Verifying session...")
        try:
            unhex_M2 = self._safe_unhexlify(M2)
        except TypeError:
            logger.error("Bad data from server (HAWK)")
            raise SRPAuthBadDataFromServer()

        self._srp_user.verify_session(unhex_M2)

        if not self._srp_user.authenticated():
            logger.error("Auth verification failed.")
            raise SRPAuthVerificationFailed()
        logger.debug("Session verified.")

        session_id = self._session.cookies.get(self.SESSION_ID_KEY, None)
        if not session_id:
            logger.error("Bad cookie from server (missing _session_id)")
            raise SRPAuthNoSessionId()

        # make the rpc call async
        emit(catalog.CLIENT_SESSION_ID, session_id)

        self.set_session_id(session_id)
        logger.debug("SUCCESS LOGIN")
        return True

    def _threader(self, cb, res, *args, **kwargs):
        return threads.deferToThread(cb, res, *args, **kwargs)

    def _change_password(self, current_password, new_password):
        """
        Changes the password for the currently logged user if the current
        password match.
        It requires to be authenticated.

        Might raise:
            SRPAuthBadUserOrPassword
            requests.exceptions.HTTPError

        :param current_password: the current password for the logged user.
        :type current_password: str
        :param new_password: the new password for the user
        :type new_password: str
        """
        leap_assert(self.get_uuid() is not None)

        if current_password != self._password:
            raise SRPAuthBadUserOrPassword

        url = "%s/%s/users/%s.json" % (
            self._provider_config.get_api_uri(),
            self._provider_config.get_api_version(),
            self.get_uuid())

        salt, verifier = self._srp.create_salted_verification_key(
            self._username.encode('utf-8'), new_password.encode('utf-8'),
            self._hashfun, self._ng)

        cookies = {self.SESSION_ID_KEY: self.get_session_id()}
        headers = {
            self.AUTHORIZATION_KEY:
            "Token token={0}".format(self.get_token())
        }
        user_data = {
            self.USER_VERIFIER_KEY: binascii.hexlify(verifier),
            self.USER_SALT_KEY: binascii.hexlify(salt)
        }

        change_password = self._session.put(
            url, data=user_data,
            verify=self._provider_config.get_ca_cert_path(),
            cookies=cookies,
            timeout=REQUEST_TIMEOUT,
            headers=headers)

        # In case of non 2xx it raises HTTPError
        change_password.raise_for_status()

        self._password = new_password

    def change_password(self, current_password, new_password):
        """
        Changes the password for the currently logged user if the current
        password match.
        It requires to be authenticated.

        :param current_password: the current password for the logged user.
        :type current_password: str
        :param new_password: the new password for the user
        :type new_password: str
        """
        d = threads.deferToThread(
            self._change_password, current_password, new_password)
        return d

    def authenticate(self, username, password):
        """
        Executes the whole authentication process for a user

        Might raise SRPAuthenticationError

        :param username: username for this session
        :type username: unicode
        :param password: password for this user
        :type password: unicode

        :returns: A defer on a different thread
        :rtype: twisted.internet.defer.Deferred
        """
        leap_assert(self.get_session_id() is None, "Already logged in")

        # User credentials stored for password changing checks
        self._username = username
        self._password = password

        self._reset_session()

        d = threads.deferToThread(self._authentication_preprocessing,
                                  username=username,
                                  password=password)
        d.addCallback(partial(self._start_authentication, username=username))

        d.addCallback(partial(self._process_challenge, username=username))
        d.addCallback(self._extract_data)
        d.addCallback(self._verify_session)
        return d

    def logout(self):
        """
        Logs out the current session.
        Expects a session_id to exists, might raise AssertionError
        """
        logger.debug("Starting logout...")

        if self.get_session_id() is None:
            logger.debug("Already logged out")
            return

        logout_url = "%s/%s/%s/" % (self._provider_config.get_api_uri(),
                                    self._provider_config.
                                    get_api_version(),
                                    "logout")
        cookies = {self.SESSION_ID_KEY: self.get_session_id()}
        headers = {
            self.AUTHORIZATION_KEY:
            "Token token={0}".format(self.get_token())
        }
        try:
            res = self._session.delete(
                logout_url,
                cookies=cookies,
                headers=headers,
                verify=self._provider_config.
                get_ca_cert_path(),
                timeout=REQUEST_TIMEOUT)
        except Exception as e:
            logger.warning("Something went wrong with the logout: %r" %
                           (e,))
            raise
        else:
            self.set_session_id(None)
            self.set_uuid(None)
            self.set_token(None)
            # Also reset the session
            self._session = self._fetcher.session()
            if res.status_code == 204:
                logger.debug("Successfully logged out.")
            else:
                logger.debug("Logout status code: %s" % res.status_code)

    def set_session_id(self, session_id):
        with self._session_id_lock:
            self._session_id = session_id

    def get_session_id(self):
        with self._session_id_lock:
            return self._session_id

    def set_uuid(self, uuid):
        with self._uuid_lock:
            full_uid = "%s@%s" % (
                self._username, self._provider_config.get_domain())
            if uuid is not None:  # avoid removing the uuid from settings
                self._settings.set_uuid(full_uid, uuid)
            self._uuid = uuid

    def get_uuid(self):
        with self._uuid_lock:
            return self._uuid

    def set_token(self, token):
        with self._token_lock:
            self._token = token

    def get_token(self):
        with self._token_lock:
            return self._token

    def is_authenticated(self):
        """
        Return whether the user is authenticated or not.

        :rtype: bool
        """
        user = self._srp_user
        if user is not None:
            return user.authenticated()

        return False


class SRPAuth(object):
    """
    SRPAuth singleton
    """
    class __impl(SRPAuthImpl):

        def __init__(self, provider_config, signaler=None):
            SRPAuthImpl.__init__(self, provider_config)
            self._signaler = signaler

        def authenticate(self, username, password):
            d = SRPAuthImpl.authenticate(self, username, password)
            d.addCallback(self._authenticate_ok)
            d.addErrback(self._authenticate_error)
            return d

        def _authenticate_ok(self, _):
            """
            Callback that notifies that the authentication was successful.

            :param _: IGNORED, output from the previous callback (None)
            :type _: IGNORED
            """
            logger.debug("Successful login!")
            self._signaler.signal(self._signaler.srp_auth_ok)

        def _authenticate_error(self, failure):
            """
            Error handler for the srpauth.authenticate method.

            :param failure: failure object that Twisted generates
            :type failure: twisted.python.failure.Failure
            """
            logger.error("Error logging in, {0!r}".format(failure))

            signal = None
            if failure.check(CancelledError):
                logger.debug("Defer cancelled.")
                failure.trap(Exception)
                return

            if self._signaler is None:
                return

            if failure.check(SRPAuthBadUserOrPassword):
                signal = self._signaler.srp_auth_bad_user_or_password
            elif failure.check(SRPAuthConnectionError):
                signal = self._signaler.srp_auth_connection_error
            elif failure.check(SRPAuthenticationError):
                signal = self._signaler.srp_auth_server_error
            else:
                signal = self._signaler.srp_auth_error

            self._signaler.signal(signal)

        def change_password(self, current_password, new_password):
            """
            Changes the password for the currently logged user if the current
            password match.
            It requires to be authenticated.

            :param current_password: the current password for the logged user.
            :type current_password: str
            :param new_password: the new password for the user
            :type new_password: str
            """
            d = SRPAuthImpl.change_password(self, current_password,
                                            new_password)
            d.addCallback(self._change_password_ok)
            d.addErrback(self._change_password_error)
            return d

        def _change_password_ok(self, _):
            """
            Password change callback.
            """
            if self._signaler is not None:
                self._signaler.signal(self._signaler.srp_password_change_ok)

        def _change_password_error(self, failure):
            """
            Password change errback.
            """
            logger.debug(
                "Error changing password. Failure: {0}".format(failure))
            if self._signaler is None:
                return

            if failure.check(SRPAuthBadUserOrPassword):
                self._signaler.signal(self._signaler.srp_password_change_badpw)
            else:
                self._signaler.signal(self._signaler.srp_password_change_error)

        def logout(self):
            try:
                SRPAuthImpl.logout(self)
                if self._signaler is not None:
                    self._signaler.signal(self._signaler.srp_logout_ok)
            except Exception:
                if self._signaler is not None:
                    self._signaler.signal(self._signaler.srp_logout_error)
                raise

    __instance = None

    def __init__(self, provider_config, signaler=None):
        """
        Create a singleton instance if needed

        :param provider_config: ProviderConfig needed to authenticate.
        :type provider_config: ProviderConfig
        :param signaler: Signaler object used to send notifications
                         from the backend
        :type signaler: Signaler
        """
        # Check whether we already have an instance
        if SRPAuth.__instance is None:
            # Create and remember instance
            SRPAuth.__instance = SRPAuth.__impl(provider_config, signaler)

        # Store instance reference as the only member in the handle
        self.__dict__['_SRPAuth__instance'] = SRPAuth.__instance

        # Generally, we initialize this with a provider_config once,
        # and after that initialize it without one and use the one
        # that was assigned before. But we need to update it if we
        # want to be able to logout and login into another provider.
        if provider_config is not None:
            SRPAuth.__instance._provider_config = provider_config

    def authenticate(self, username, password):
        """
        Executes the whole authentication process for a user

        Might raise SRPAuthenticationError based

        :param username: username for this session
        :type username: str
        :param password: password for this user
        :type password: str
        :returns: a Deferred that will fire when the authentication is done
        :rtype: Deferred
        """
        username = username.lower()
        d = self.__instance.authenticate(username, password)
        return d

    def is_authenticated(self):
        """
        Return whether the user is authenticated or not.

        :rtype: bool
        """
        return self.__instance.is_authenticated()

    def change_password(self, current_password, new_password):
        """
        Changes the user's password.

        :param current_password: the current password of the user.
        :type current_password: str
        :param new_password: the new password for the user.
        :type new_password: str

        :returns: a defer to interact with.
        :rtype: twisted.internet.defer.Deferred
        """
        d = self.__instance.change_password(current_password, new_password)
        return d

    def get_username(self):
        """
        Returns the username of the currently authenticated user or None if
        no user is logged.

        :rtype: str or None
        """
        if self.get_uuid() is None:
            return None
        return self.__instance._username

    def get_session_id(self):
        return self.__instance.get_session_id()

    def get_uuid(self):
        return self.__instance.get_uuid()

    def get_token(self):
        return self.__instance.get_token()

    def logout(self):
        """
        Logs out the current session.
        Expects a session_id to exists, might raise AssertionError
        """
        try:
            self.__instance.logout()
            logger.debug("Logout success")
            return True
        except Exception as e:
            logger.debug("Logout error: {0!r}".format(e))
        return False



# -*- coding: utf-8 -*-
# systray.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Custom system tray manager.
"""

from PySide import QtGui

from leap.common.check import leap_assert_type


class SysTray(QtGui.QSystemTrayIcon):
    """
    Custom system tray that allows us to use a more 'intelligent' tooltip.
    """

    def __init__(self, parent=None):
        QtGui.QSystemTrayIcon.__init__(self, parent)
        self._services = {}

    def set_service_tooltip(self, service, tooltip):
        """
        Sets the service tooltip.

        :param service: service name identifier
        :type service: unicode
        :param tooltip: tooltip to display for that service
        :type tooltip: unicode
        """
        leap_assert_type(service, unicode)
        leap_assert_type(tooltip, unicode)

        self._services[service] = tooltip
        tooltip = "\n".join(self._services.values())
        self.setToolTip(tooltip)



# -*- coding: utf-8 -*-
# keyring_helpers.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Keyring helpers.
"""
try:
    import keyring
    from keyring.backends.file import EncryptedKeyring, PlaintextKeyring
    OBSOLETE_KEYRINGS = [
        EncryptedKeyring,
        PlaintextKeyring
    ]
    canuse = lambda kr: (kr is not None
                         and kr.__class__ not in OBSOLETE_KEYRINGS)

except Exception:
    # Problems when importing keyring! It might be a problem binding to the
    # dbus socket, or stuff like that.
    keyring = None


from leap.bitmask.logs.utils import get_logger
logger = get_logger()


def _get_keyring_with_fallback():
    """
    Get the default keyring, and if obsolete try to pick SecretService keyring
    if available.

    This is a workaround for the cases in which the keyring module chooses
    an insecure keyring by default (ie, inside a virtualenv).
    """
    if keyring is None:
        return None
    kr = keyring.get_keyring()
    if not canuse(kr):
        logger.debug("No usable keyring found")
        return None
    logger.debug("Selected keyring: %s" % (kr.__class__,))
    if not canuse(kr):
        logger.debug("Not using default keyring since it is obsolete")
    return kr


def has_keyring():
    """
    Return whether we have an useful keyring to use.

    :rtype: bool
    """
    if keyring is None:
        return False
    kr = _get_keyring_with_fallback()
    return canuse(kr)


def get_keyring():
    """
    Return an usable keyring.

    :rtype: keyringBackend or None
    """
    if keyring is None:
        return False
    kr = _get_keyring_with_fallback()
    return kr if canuse(kr) else None



# -*- coding: utf-8 -*-
# credentials.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Credentials utilities
"""
from PySide import QtCore, QtGui

WEAK_PASSWORDS = ("123456", "qweasd", "qwerty", "password")
USERNAME_REGEX = r"^[a-z][a-z\d_\-\.]+[a-z\d]$"

USERNAME_VALIDATOR = QtGui.QRegExpValidator(QtCore.QRegExp(USERNAME_REGEX))


def username_checks(username):
    # translation helper
    _tr = QtCore.QObject().tr

    message = None

    if message is None and len(username) < 2:
        message = _tr("Username must have at least 2 characters")

    valid = USERNAME_VALIDATOR.validate(username, 0)
    valid_username = valid[0] == QtGui.QValidator.State.Acceptable
    if message is None and not valid_username:
        message = _tr("That username is not allowed. Try another.")

    return message is None, message


def password_checks(username, password, password2):
    """
    Performs basic password checks to avoid really easy passwords.

    :param username: username provided at the registrarion form
    :type username: str
    :param password: password from the registration form
    :type password: str
    :param password2: second password from the registration form
    :type password: str

    :returns: (True, None, None) if all the checks pass,
              (False, message, field name) otherwise
    :rtype: tuple(bool, str, str)
    """
    # translation helper
    _tr = QtCore.QObject().tr

    message = None
    field = None

    if message is None and password != password2:
        message = _tr("Passwords don't match")
        field = 'new_password_confirmation'

    if message is None and not password:
        message = _tr("Password is empty")
        field = 'new_password'

    if message is None and len(password) < 8:
        message = _tr("Password is too short")
        field = 'new_password'

    if message is None and password in WEAK_PASSWORDS:
        message = _tr("Password is too easy")
        field = 'new_password'

    if message is None and username == password:
        message = _tr("Password can't be the same as username")
        field = 'new_password'

    return message is None, message, field



# -*- coding: utf-8 -*-
# initializers.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Platform-dependant initialization code.
"""
import os
import platform
import stat
import subprocess
import tempfile

from PySide import QtGui, QtCore

from leap.bitmask.config import flags
from leap.bitmask.config.leapsettings import LeapSettings
from leap.bitmask.services.eip import get_vpn_launcher
from leap.bitmask.services.eip.darwinvpnlauncher import DarwinVPNLauncher
from leap.bitmask.util import first
from leap.bitmask.util.privilege_policies import LinuxPolicyChecker
from leap.bitmask.logs.utils import get_logger

logger = get_logger()

# NOTE we could use a deferToThread here, but should
# be aware of this bug: http://www.themacaque.com/?p=1067

__all__ = ["init_platform"]

_system = platform.system()


class InitSignals(QtCore.QObject):
    """
    Signal container to communicate initialization events to differnt widgets.
    """
    eip_missing_helpers = QtCore.Signal()


init_signals = InitSignals()


def init_platform():
    """
    Run the right initializer for the platform we are running in.

    :return: whether the initializacion succeeded or not.
    :rtype: bool
    """
    initializer = None
    ok = False
    try:
        initializer = globals()[_system + "Initializer"]
    except:
        pass
    if initializer:
        logger.debug("Running initializer for %s" % (_system, ))
        ok = initializer()
    else:
        logger.debug("Initializer not found for %s" % (_system, ))

    return ok


#
# common utils
#

NOTFOUND_MSG = ("Tried to install %s, but %s "
                "not found inside this bundle.")
BADEXEC_MSG = ("Tried to install %s, but %s "
               "failed to %s.")

HELPERS_NOTFOUND_MSG = NOTFOUND_MSG % (
    "helper files", "those were")
HELPERS_BADEXEC_MSG = BADEXEC_MSG % (
    "helper files", "they", "be copied")


def get_missing_helpers_dialog():
    """
    Create a dialog for notifying of missing helpers.
    Returns that dialog.

    :rtype: QtGui.QMessageBox instance
    """
    WE_NEED_POWERS = ("To better protect your privacy, "
                      "Bitmask needs administrative privileges "
                      "to install helper files. Encrypted "
                      "Internet cannot work without those files. "
                      "Do you want to install them now?")
    msg = QtGui.QMessageBox()
    msg.setWindowTitle(msg.tr("Missing helper files"))
    msg.setText(msg.tr(WE_NEED_POWERS))
    # but maybe the user really deserve to know more
    # msg.setInformativeText(msg.tr(BECAUSE))
    msg.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
    msg.addButton("No, don't ask again", QtGui.QMessageBox.RejectRole)
    msg.setDefaultButton(QtGui.QMessageBox.Yes)
    msg.setEscapeButton(QtGui.QMessageBox.No)
    return msg


def check_missing():
    """
    Check for the need of installing missing scripts, and
    raises a dialog to ask user for permission to do it.
    """
    config = LeapSettings()
    complain_missing = False
    alert_missing = config.get_alert_missing_scripts()

    if alert_missing and not flags.STANDALONE:
        # We refuse to install missing stuff if not running with standalone
        # flag. Right now we rely on the flag alone, but we can disable this
        # by overwriting some constant from within the debian package.
        alert_missing = False
        complain_missing = True

    launcher = get_vpn_launcher()
    missing_scripts = launcher.missing_updown_scripts()
    missing_other = launcher.missing_other_files()

    if missing_scripts:
        logger.warning("Missing scripts: %s" % (missing_scripts))
    if missing_other:
        logger.warning("Missing other files: %s" % (missing_other))

    missing_some = missing_scripts or missing_other
    if alert_missing and missing_some:
        msg = get_missing_helpers_dialog()
        ret = msg.exec_()

        if ret == QtGui.QMessageBox.Yes:
            install_missing_fun = globals().get(
                "_%s_install_missing_scripts" % (_system.lower(),),
                None)
            if not install_missing_fun:
                logger.warning(
                    "Installer not found for platform %s." % (_system,))
                return

            # XXX maybe move constants to fun
            ok = install_missing_fun(HELPERS_BADEXEC_MSG, HELPERS_NOTFOUND_MSG)
            if not ok:
                msg = QtGui.QMessageBox()
                msg.setWindowTitle(msg.tr("Problem installing files"))
                msg.setText(msg.tr('Some of the files could not be copied.'))
                msg.setIcon(QtGui.QMessageBox.Warning)
                msg.exec_()

        elif ret == QtGui.QMessageBox.No:
            logger.debug("Not installing missing scripts, "
                         "user decided to ignore our warning.")
            init_signals.eip_missing_helpers.emit()

        elif ret == QtGui.QMessageBox.Rejected:
            logger.debug(
                "Setting alert_missing_scripts to False, we will not "
                "ask again")
            init_signals.eip_missing_helpers.emit()
            config.set_alert_missing_scripts(False)

    if complain_missing and missing_some:
        missing = missing_scripts + missing_other
        msg = _get_missing_complain_dialog(missing)
        ret = msg.exec_()

    # If there is some missing file and we don't want to complain, we emit the
    # 'missing helpers' signal so the eip status can show that some files are
    # missing.
    if missing_some and not alert_missing and not complain_missing:
        init_signals.eip_missing_helpers.emit()


def check_polkit():
    """
    Check if we have a running polkit agent and tries to launch an agent if
    needed.
    Show an error message if there is no agent and we couldn't run one.

    :return: True if we have a polkit running (or if we started one), False
             otherwise
    :rtype: bool
    """
    if LinuxPolicyChecker.is_up():
        return True

    try:
        LinuxPolicyChecker.maybe_pkexec()
        return True
    except Exception:
        logger.error("No polkit agent running.")

        msg = QtGui.QMessageBox()
        msg.setWindowTitle(msg.tr("No polkit agent running"))
        msg.setText(
            msg.tr('There is no polkit agent running and it is needed to run '
                   'the Bitmask services.<br>Take a look at the '
                   '<a href="https://leap.se/en/docs/client/known-issues">'
                   'known issues</a> page'))
        msg.setIcon(QtGui.QMessageBox.Critical)
        msg.exec_()

        return False


#
# windows initializers
#


def _windows_has_tap_device():
    """
    Loop over the windows registry trying to find if the tap0901 tap driver
    has been installed on this machine.
    """
    import _winreg as reg

    adapter_key = 'SYSTEM\CurrentControlSet\Control\Class' \
        '\{4D36E972-E325-11CE-BFC1-08002BE10318}'
    with reg.OpenKey(reg.HKEY_LOCAL_MACHINE, adapter_key) as adapters:
        try:
            for i in xrange(10000):
                key_name = reg.EnumKey(adapters, i)
                with reg.OpenKey(adapters, key_name) as adapter:
                    try:
                        component_id = reg.QueryValueEx(adapter,
                                                        'ComponentId')[0]
                        if component_id.startswith("tap0901"):
                            return True
                    except WindowsError:
                        pass
        except WindowsError:
            pass
    return False


def WindowsInitializer():
    """
    Raise a dialog in case that the windows tap driver has not been found
    in the registry, asking the user for permission to install the driver
    """
    if not _windows_has_tap_device():
        msg = QtGui.QMessageBox()
        msg.setWindowTitle(msg.tr("TAP Driver"))
        msg.setText(msg.tr("Bitmask needs to install the necessary drivers "
                           "for Encrypted Internet to work. Would you like to "
                           "proceed?"))
        msg.setInformativeText(msg.tr("Encrypted Internet uses VPN, which "
                                      "needs a TAP device installed and none "
                                      "has been found. This will ask for "
                                      "administrative privileges."))
        msg.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
        msg.setDefaultButton(QtGui.QMessageBox.Yes)
        ret = msg.exec_()

        if ret == QtGui.QMessageBox.Yes:
            # XXX should do this only if executed inside bundle.
            # Let's assume it's the only way it's gonna be executed under win
            # by now.
            driver_path = os.path.join(os.getcwd(),
                                       "apps",
                                       "eip",
                                       "tap_driver")
            dev_installer = os.path.join(driver_path,
                                         "devcon.exe")
            if os.path.isfile(dev_installer) and \
                    stat.S_IXUSR & os.stat(dev_installer)[stat.ST_MODE] != 0:
                inf_path = os.path.join(driver_path,
                                        "OemWin2k.inf")
                cmd = [dev_installer, "install", inf_path, "tap0901"]
                ret = subprocess.call(cmd, stdout=subprocess.PIPE, shell=True)
            else:
                logger.error("Tried to install TAP driver, but the installer "
                             "is not found or not executable")

#
# Darwin initializer functions
#


def _darwin_has_tun_kext():
    """
    Return True only if we found a directory under the system kext folder
    containing a kext named tun.kext, AND we found a startup item named 'tun'
    """
    # XXX we should be smarter here and use kextstats output.

    has_kext = os.path.isdir("/Library/Extensions/tun.kext")
    has_startup = os.path.isdir("/Library/StartupItems/tun")
    has_tun_and_startup = has_kext and has_startup
    logger.debug(
        'platform initializer check: has tun_and_startup = %s' %
        (has_tun_and_startup,))
    return has_tun_and_startup


def _darwin_install_missing_scripts(badexec, notfound):
    """
    Try to install the missing up/down scripts.

    :param badexec: error for notifying execution error during command.
    :type badexec: str
    :param notfound: error for notifying missing path.
    :type notfound: str
    :returns: True if the files could be copied successfully.
    :rtype: bool
    """
    # We expect to execute this from some way of bundle, since
    # the up/down scripts should be put in place by the installer.
    success = False
    installer_path = os.path.abspath(
        os.path.join(
            os.getcwd(), "..", "Resources", "openvpn"))
    launcher = DarwinVPNLauncher

    # XXX FIXME !!! call the bash script!
    if os.path.isdir(installer_path):
        fd, tempscript = tempfile.mkstemp(prefix="leap_installer-")
        try:
            scriptlines = launcher.cmd_for_missing_scripts(installer_path)
            with os.fdopen(fd, 'w') as f:
                f.write(scriptlines)
            st = os.stat(tempscript)
            os.chmod(tempscript, st.st_mode | stat.S_IEXEC | stat.S_IXUSR |
                     stat.S_IXGRP | stat.S_IXOTH)

            cmd, args = launcher().get_cocoasudo_installmissing_cmd()
            args.append(tempscript)
            cmdline = " ".join([cmd] + args)
            ret = subprocess.call(
                cmdline, stdout=subprocess.PIPE,
                shell=True)
            success = ret == 0
            if not success:
                logger.error("Install missing scripts failed.")
        except Exception as exc:
            logger.error(badexec)
            logger.error("Error was: %r" % (exc,))
        finally:
            try:
                os.remove(tempscript)
            except OSError as exc:
                logger.error("%r" % (exc,))
    else:
        logger.error(notfound)
        logger.debug('path searched: %s' % (installer_path,))

    return success


def DarwinInitializer():
    """
    Raise a dialog in case that the osx tuntap driver has not been found
    in the registry, asking the user for permission to install the driver
    """
    # XXX split this function into several

    TUNTAP_NOTFOUND_MSG = NOTFOUND_MSG % (
        "tuntaposx kext", "the installer")
    TUNTAP_BADEXEC_MSG = BADEXEC_MSG % (
        "tuntaposx kext", "the installer", "be launched")

    # TODO DRY this with other cases, and
    # factor out to _should_install() function.
    # Leave the dialog as a more generic thing.

    if not _darwin_has_tun_kext():
        msg = QtGui.QMessageBox()
        msg.setWindowTitle(msg.tr("TUN Driver"))
        msg.setText(msg.tr("Bitmask needs to install the necessary drivers "
                           "for Encrypted Internet to work. Would you like to "
                           "proceed?"))
        msg.setInformativeText(msg.tr("Encrypted Internet uses VPN, which "
                                      "needs a kernel extension for a TUN "
                                      "device installed, and none "
                                      "has been found. This will ask for "
                                      "administrative privileges."))
        msg.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
        msg.setDefaultButton(QtGui.QMessageBox.Yes)
        ret = msg.exec_()

        if ret == QtGui.QMessageBox.Yes:
            installer_path = os.path.abspath(
                os.path.join(
                    os.getcwd(),
                    "..",
                    "Resources",
                    "tuntap-installer.app"))
            if os.path.isdir(installer_path):
                cmd = ["open '%s'" % (installer_path,)]
                try:
                    ret = subprocess.call(
                        cmd, stdout=subprocess.PIPE,
                        shell=True)
                except:
                    logger.error(TUNTAP_BADEXEC_MSG)
            else:
                logger.error(TUNTAP_NOTFOUND_MSG)

    # Second check, for missing scripts.
    check_missing()
    return True


#
# Linux initializers
#


def _get_missing_complain_dialog(stuff):
    """
    Create a dialog for notifying about missing helpers (but doing nothing).
    Used from non-standalone runs.

    :param stuff: list of missing items to display
    :type stuff: list
    :rtype: QtGui.QMessageBox instance
    """
    msgstr = QtCore.QObject()
    msgstr.NO_HELPERS = msgstr.tr(
        "Some essential helper files are missing in your system.")
    msgstr.EXPLAIN = msgstr.tr(
        "Reinstall your debian packages, or make sure you place them by hand.")

    class ComplainDialog(QtGui.QDialog):

        def __init__(self, parent=None):
                super(ComplainDialog, self).__init__(parent)

                label = QtGui.QLabel(msgstr.NO_HELPERS)
                label.setAlignment(QtCore.Qt.AlignLeft)

                label2 = QtGui.QLabel(msgstr.EXPLAIN)
                label2.setAlignment(QtCore.Qt.AlignLeft)

                textedit = QtGui.QTextEdit()
                textedit.setText("\n".join(stuff))

                ok = QtGui.QPushButton()
                ok.setText(self.tr("Ok, thanks"))
                self.ok = ok
                self.ok.clicked.connect(self.close)

                mainLayout = QtGui.QGridLayout()
                mainLayout.addWidget(label, 0, 0)
                mainLayout.addWidget(label2, 1, 0)
                mainLayout.addWidget(textedit, 2, 0)
                mainLayout.addWidget(ok, 3, 0)

                self.setLayout(mainLayout)

    msg = ComplainDialog()
    msg.setWindowTitle(msg.tr("Missing Bitmask helpers"))
    return msg


def _linux_install_missing_scripts(badexec, notfound):
    """
    Try to install the missing helper files.

    :param badexec: error for notifying execution error during command.
    :type badexec: str
    :param notfound: error for notifying missing path.
    :type notfound: str
    :returns: True if the files could be copied successfully.
    :rtype: bool
    """
    success = False
    installer_path = os.path.abspath(
        os.path.join(os.getcwd(), "apps", "eip", "files"))

    install_helper = "leap-install-helper.sh"
    install_helper_path = os.path.join(installer_path, install_helper)

    install_opts = ("--from-path %s --install-bitmask-root YES "
                    "--install-polkit-file YES --install-openvpn YES "
                    "--remove-old-files YES" % (installer_path,))

    if os.path.isdir(installer_path):
        try:
            policyChecker = LinuxPolicyChecker()
            pkexec = first(policyChecker.maybe_pkexec())
            cmdline = ["%s %s %s" % (
                pkexec, install_helper_path, install_opts)]

            ret = subprocess.call(
                cmdline, stdout=subprocess.PIPE,
                shell=True)
            success = ret == 0
            if not success:
                logger.error("Install of helpers failed.")
        except Exception as exc:
            logger.error(badexec)
            logger.error("Error was: %r" % (exc,))
    else:
        logger.error(notfound)
        logger.debug('path searched: %s' % (installer_path,))

    return success


def LinuxInitializer():
    """
    Raise a dialog if needed files are missing.

    Missing files can be either bitmask-root policykit file.
    The dialog will also be raised if some of those files are
    found to have incorrect permissions.

    :return: whether the operations went ok or not, if False, it's recommended
    not to continue with the app run.
    :rtype: bool
    """
    if not check_polkit():
        return False

    check_missing()
    return True



# -*- coding: utf-8 -*-
# login.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Login widget implementation

The login sequence is the following:
    - _do_login
    - backend.provider_setup (
        check_name_resolution, check_https, download_provider_info)
        - on error:   _provider_setup_intermediate
        - on success: _load_provider_config
    - backend.provider_bootstrap (
        download_ca_cert, check_ca_fingerprint, check_api_certificate)
        - on error:   _provider_setup_intermediate
        - on success: _provider_config_loaded
    - backend.user_login
        - on error:   _authentication_error
        - on success: _authentication_finished

"""
from keyring.errors import InitError as KeyringInitError
from PySide import QtCore, QtGui
from ui_login import Ui_LoginWidget

# TODO: we should use a more granular signaling instead of passing error/ok as
# a result.
from leap.bitmask.backend.leapbackend import ERROR_KEY, PASSED_KEY
from leap.bitmask.config import flags
from leap.bitmask.config.leapsettings import LeapSettings
from leap.bitmask.logs.utils import get_logger
from leap.bitmask.gui.signaltracker import SignalTracker
from leap.bitmask.util import make_address
from leap.bitmask.util.credentials import USERNAME_REGEX
from leap.bitmask.util.keyring_helpers import has_keyring
from leap.bitmask.util.keyring_helpers import get_keyring
from leap.common.check import leap_assert_type

logger = get_logger()


class LoginState(object):
    """
    This class holds the states related to the login sequence.
    """

    def __init__(self):
        # `wait_to_login` defines whether we should wait to start the login
        # sequence or we should start right away.
        # This is used, for instance, to hold on until EIP is started since the
        # firewall could block the login attempt.
        self.wait_to_login = False

        # This state indicates that the login sequence was required to start
        # but it was set on hold since we `wait_to_login` was True
        self.login_waiting = False

        # Full username of the logged user, with the format: 'user@provider'
        self.full_logged_username = None


class LoginWidget(QtGui.QWidget, SignalTracker):
    """
    Login widget that emits signals to display the wizard or to
    perform login.
    """
    login_start = QtCore.Signal()
    login_finished = QtCore.Signal()
    login_offline_finished = QtCore.Signal()
    login_failed = QtCore.Signal()
    logged_out = QtCore.Signal()

    MAX_STATUS_WIDTH = 40

    # Keyring
    KEYRING_KEY = "bitmask"

    def __init__(self, backend, signaler, parent=None):
        """
        Constructs the LoginWidget.

        :param backend: Backend being used
        :type backend: Backend
        :param signaler: Object in charge of handling communication
                         back to the frontend
        :type signaler: Signaler
        :param parent: The parent widget for this widget
        :type parent: QWidget or None
        """
        QtGui.QWidget.__init__(self, parent)
        SignalTracker.__init__(self)

        self.ui = Ui_LoginWidget()
        self.ui.setupUi(self)

        self.ui.chkRemember.stateChanged.connect(self._remember_state_changed)
        self.ui.chkRemember.setEnabled(has_keyring())

        self.ui.lnUser.textChanged.connect(self._credentials_changed)
        self.ui.lnPassword.textChanged.connect(self._credentials_changed)

        self.ui.btnLogin.clicked.connect(self._do_login)
        self.ui.btnLogout.clicked.connect(self.do_logout)

        self.ui.lnUser.setValidator(
            QtGui.QRegExpValidator(QtCore.QRegExp(USERNAME_REGEX), self))

        self.ui.clblErrorMsg.hide()
        self.ui.clblErrorMsg.clicked.connect(self.ui.clblErrorMsg.hide)

        self.ui.lnUser.textEdited.connect(self.ui.clblErrorMsg.hide)
        self.ui.lnPassword.textEdited.connect(self.ui.clblErrorMsg.hide)

        self._settings = LeapSettings()
        self._backend = backend
        self._leap_signaler = signaler

        # the selected provider that we'll use to login
        self._provider = None

        self._state = LoginState()

        self._set_logged_out()

    def _remember_state_changed(self, state):
        """
        Save the remember state in the LeapSettings.

        :param state: the current state of the check box.
        :type state: int
        """
        # The possible state values of the checkbox (from QtCore.Qt.CheckState)
        # are: Checked, Unchecked and PartiallyChecked
        self._settings.set_remember(state == QtCore.Qt.Checked)

    def _credentials_changed(self, text):
        """
        TRIGGER:
            self.ui.lnUser.textChanged
            self.ui.lnPassword.textChanged

        Update the 'enabled' status of the login button depending if we have
        all the fields needed set.
        """
        enabled = self._provider and self.get_user() and self.get_password()
        enabled = bool(enabled)  # provider can be None

        self.ui.btnLogin.setEnabled(enabled)

    def wait_for_login(self, wait):
        """
        Set the wait flag to True/False so the next time that a login action is
        requested it will wait or not.

        If we set the wait to True and we have paused a login request before,
        this will trigger a login action.

        :param wait: whether we should wait or not on the next login request.
        :type wait: bool
        """
        self._state.wait_to_login = wait

        if not wait and self._state.login_waiting:
            logger.debug("No more waiting, triggering login sequence.")
            self._do_login()

    def set_provider(self, provider):
        """
        Set the provider to use in the login sequence.

        :param provider: the provider to use.
        :type provider: unicode
        """
        self._provider = provider

    def set_remember(self, value):
        """
        Check the remember user and password checkbox

        :param value: True to mark it checked, False otherwise
        :type value: bool
        """
        self.ui.chkRemember.setChecked(value)

    def get_remember(self):
        """
        Returns the remember checkbox state

        :rtype: bool
        """
        return self.ui.chkRemember.isChecked()

    def set_user(self, user):
        """
        Sets the user and focuses on the next field, password.

        :param user: user to set the field to
        :type user: str
        """
        self.ui.lnUser.setText(user)
        self._focus_password()

    def get_user(self):
        """
        Return the user that appears in the widget.

        :rtype: unicode
        """
        return self.ui.lnUser.text()

    def get_logged_user(self):
        """
        Return the current logged user or None if no user is logged in.
        The return value has the format: 'user@provider'

        :rtype: unicode or None
        """
        return self._state.full_logged_username

    def set_password(self, password):
        """
        Sets the password for the widget

        :param password: password to set
        :type password: unicode
        """
        self.ui.lnPassword.setText(password)

    def get_password(self):
        """
        Returns the password that appears in the widget

        :rtype: unicode
        """
        return self.ui.lnPassword.text()

    def set_status(self, status, error=True):
        """
        Sets the status label at the login stage to status

        :param status: status message
        :type status: str
        """
        if len(status) > self.MAX_STATUS_WIDTH:
            status = status[:self.MAX_STATUS_WIDTH] + "..."
        if error:
            self.ui.clblErrorMsg.show()
            self.ui.clblErrorMsg.setText(status)
        else:
            self.ui.lblStatus.setText(status)

    def set_enabled(self, enabled=False):
        """
        Enables or disables all the login widgets

        :param enabled: wether they should be enabled or not
        :type enabled: bool
        """
        self.ui.lnUser.setEnabled(enabled)
        self.ui.lnPassword.setEnabled(enabled)
        self.ui.chkRemember.setEnabled(enabled and has_keyring())

        self._set_cancel(not enabled)

    def set_logout_btn_enabled(self, enabled):
        """
        Enables or disables the logout button.

        :param enabled: wether they should be enabled or not
        :type enabled: bool
        """
        self.ui.btnLogout.setEnabled(enabled)

    def _set_cancel(self, enabled=False):
        """
        Enables or disables the cancel action in the "log in" process.

        :param enabled: wether it should be enabled or not
        :type enabled: bool
        """
        text = self.tr("Cancel")
        login_or_cancel = self._do_cancel
        hide_remember = enabled

        if not enabled:
            text = self.tr("Log In")
            login_or_cancel = self._do_login

        self.ui.btnLogin.setText(text)

        self.ui.btnLogin.clicked.disconnect()
        self.ui.btnLogin.clicked.connect(login_or_cancel)
        self.ui.chkRemember.setVisible(not hide_remember)
        self.ui.lblStatus.setVisible(hide_remember)

    def _focus_password(self):
        """
        Focuses in the password lineedit
        """
        self.ui.lnPassword.setFocus()

    def _check_login(self):
        """
        Check that we have the needed fields to do the actual login: provider,
        username and password.

        :return: True if everything's good to go, False otherwise.
        :rtype: bool
        """
        provider = self._provider
        username = self.get_user()
        password = self.get_password()

        if not provider:
            self.set_status(self.tr("Please select a valid provider"))
            return False

        if not username:
            self.set_status(self.tr("Please provide a valid username"))
            return False

        if not password:
            self.set_status(self.tr("Please provide a valid password"))
            return False

        return True

    def _set_logging_in(self):
        """
        Set the status of the widget to "Logging in".
        """
        self.set_status(self.tr("Logging in..."), error=False)
        self.set_enabled(False)
        self.ui.clblErrorMsg.hide()

    def _save_credentials(self):
        """
        If the user asked to remember the credentials, we save them into the
        keyring.
        """
        provider = self._provider
        username = self.get_user()
        password = self.get_password()

        self._settings.set_provider(provider)
        if self.get_remember() and has_keyring():
            # in the keyring and in the settings
            # we store the value 'usename@provider'
            full_user_id = make_address(username, provider).encode("utf8")
            try:
                keyring = get_keyring()
                keyring.set_password(self.KEYRING_KEY,
                                     full_user_id, password.encode("utf8"))
                # Only save the username if it was saved correctly in
                # the keyring
                self._settings.set_user(full_user_id)
            except KeyringInitError as e:
                logger.error("Failed to unlock keyring, maybe the user "
                             "cancelled the operation {0!r}".format(e))
            except Exception as e:
                logger.exception("Problem saving data to keyring. %r" % (e,))

    def do_login(self):
        """
        Start the login sequence.
        We check that we have the needed fields to do the actual login:
        provider, username and password.
        If everything is ok we perform the login.

        Note that the actual login won't be started if you set the
        `wait_to_login` flag, it will be scheduled to get started when you set
        that flag to False.

        :return: True if the login sequence started, False otherwise.
        :rtype: bool
        """
        ok = self._provider and self.get_user() and self.get_password()

        if ok:
            self._do_login()

        return bool(ok)

    def _do_login(self):
        """
        Start the login sequence.
        """
        if self._state.wait_to_login:
            logger.debug("Login delayed, waiting...")

            self._state.login_waiting = True
            self.ui.btnLogin.setEnabled(False)
            self.ui.btnLogin.setText(self.tr("Waiting..."))
            # explicitly process events to display the button's text change.
            QtCore.QCoreApplication.processEvents(0, 10)

            return
        else:
            self._state.login_waiting = False
            self.ui.btnLogin.setEnabled(True)

        self.login_start.emit()

        provider = self._provider
        if flags.OFFLINE:
            self._do_offline_login()
            return

        # connect to the backend signals, remember to disconnect after login.
        self._backend_connect()

        if self._check_login():
            self._set_logging_in()
            self._save_credentials()
            self._backend.provider_setup(provider=provider)

    def _do_cancel(self):
        logger.debug("Cancelling log in.")

        self._backend.provider_cancel_setup()
        self._backend.user_cancel_login()
        self._set_logged_out()

    def _set_login_cancelled(self):
        """
        TRIGGERS:
            Signaler.prov_cancelled_setup

        Re-enable the login widget and display a message for the cancelled
        operation.
        """
        self.set_status(self.tr("Log in cancelled by the user."))
        self.set_enabled(True)

    def _provider_setup_intermediate(self, data):
        """
        TRIGGERS:
            self._backend.signaler.prov_name_resolution
            self._backend.signaler.prov_https_connection

        Handle a possible problem during the provider setup process.
        If there was a problem, display it, otherwise it does nothing.
        """
        if not data[PASSED_KEY]:
            logger.error(data[ERROR_KEY])
            self._login_problem_provider()

    def _login_problem_provider(self):
        """
        Warn the user about a problem with the provider during login.
        """
        self.set_status(self.tr("Unable to login: Problem with provider"))
        self.set_enabled(True)

    def _load_provider_config(self, data):
        """
        TRIGGERS:
            self._backend.signaler.prov_download_provider_info

        Once the provider config has been downloaded, start the second
        part of the bootstrapping sequence.

        :param data: result from the last stage of the
                     backend.provider_setup()
        :type data: dict
        """
        if not data[PASSED_KEY]:
            logger.error(data[ERROR_KEY])
            self._login_problem_provider()
            return

        self._backend.provider_bootstrap(provider=self._provider)

    def _provider_config_loaded(self, data):
        """
        TRIGGERS:
            self._backend.signaler.prov_check_api_certificate

        Once the provider configuration is loaded, this starts the SRP
        authentication
        """
        if not data[PASSED_KEY]:
            logger.error(data[ERROR_KEY])
            self._login_problem_provider()
            return

        self._backend.user_login(provider=self._provider,
                                 username=self.get_user(),
                                 password=self.get_password())

    # TODO check this!
    def _do_offline_login(self):
        logger.debug("OFFLINE mode! bypassing remote login")
        # TODO reminder, we're not handling logout for offline mode.
        self._set_logged_in()
        self._logged_in_offline = True
        self._set_label_offline()
        self.login_offline_finished.emit()

    def _set_label_offline(self):
        """
        Set the login label to reflect offline status.
        """
        # TODO: figure out what widget to use for this. Maybe the window title?

    def _set_logged_in(self):
        """
        Set the widgets to the logged in state.
        """
        fullname = make_address(self.get_user(), self._provider)
        self._state.full_logged_username = fullname
        self.ui.login_widget.hide()
        self.ui.logged_widget.show()
        self.ui.lblUser.setText(fullname)

    def _authentication_finished(self):
        """
        TRIGGERS:
            self._srp_auth.authentication_finished

        The SRP auth was successful, set the login status.
        """
        self.set_status(self.tr("Succeeded"), error=False)
        self._set_logged_in()
        self.disconnect_and_untrack()

        if not flags.OFFLINE:
            self.login_finished.emit()

    def _authentication_error(self, msg):
        """
        TRIGGERS:
            Signaler.srp_auth_error
            Signaler.srp_auth_server_error
            Signaler.srp_auth_connection_error
            Signaler.srp_auth_bad_user_or_password

        Handle the authentication errors.

        :param msg: the message to show to the user.
        :type msg: unicode
        """
        self.set_status(msg)
        self.set_enabled(True)
        self.login_failed.emit()

    def _set_logged_out(self):
        """
        Set the widgets to the logged out state.
        """
        # TODO consider "logging out offline" too... how that would be ???
        self._state.full_logged_username = None

        self.ui.login_widget.show()
        self.ui.logged_widget.hide()

        self.set_password("")
        self.set_enabled(True)
        self.set_status("", error=False)

    def do_logout(self):
        """
        TRIGGER:
            self.ui.btnLogout.clicked

        Start the logout sequence and set the widgets to the "logging out"
        state.
        """
        if self._state.full_logged_username is not None:
            self._set_logging_out()
            self._backend.user_logout()
        else:
            logger.debug("Not logged in.")

    def _set_logging_out(self, logging_out=True):
        """
        Set the status of the logout button.

        logging_out == True:
            button text -> "Logging out..."
            button enabled -> False

        logging_out == False:
            button text -> "Logout
            button enabled -> True

        :param logging_out: wether we are logging out or not.
        :type logging_out: bool
        """
        if logging_out:
            self.ui.btnLogout.setText(self.tr("Logging out..."))
            self.ui.btnLogout.setEnabled(False)
        else:
            self.ui.btnLogout.setText(self.tr("Logout"))
            self.ui.btnLogout.setEnabled(True)
            self.ui.clblErrorMsg.hide()

    def _logout_error(self):
        """
        TRIGGER:
            self._srp_auth.logout_error

        Inform the user about a logout error.
        """
        self._set_logging_out(False)
        self.set_status(self.tr("Something went wrong with the logout."))

    def _logout_ok(self):
        """
        TRIGGER:
            self._srp_auth.logout_ok

        Switch the stackedWidget back to the login stage after logging out.
        """
        self._set_logging_out(False)
        self._set_logged_out()
        self.logged_out.emit()

    def load_user_from_keyring(self, saved_user):
        """
        Try to load a user from the keyring.

        :param saved_user: the saved username as user@domain
        :type saved_user: unicode

        :return: True if the user was loaded successfully, False otherwise.
        :rtype: bool
        """
        leap_assert_type(saved_user, unicode)

        try:
            username, domain = saved_user.split('@')
        except ValueError as e:
            # if the saved_user does not contain an '@'
            logger.error('Username@provider malformed. %r' % (e, ))
            return False

        self.set_user(username)
        self.set_remember(True)

        saved_password = None
        try:
            keyring = get_keyring()
            u_user = saved_user.encode("utf8")
            saved_password = keyring.get_password(self.KEYRING_KEY, u_user)
        except ValueError as e:
            logger.debug("Incorrect Password. %r." % (e,))
        except KeyringInitError as e:
            logger.error("Failed to unlock keyring, maybe the user "
                         "cancelled the operation {0!r}".format(e))

        if saved_password is not None:
            self.set_password(saved_password)
            return True

        return False

    def _backend_connect(self):
        """
        Connect to backend signals.

        We track the signals in order to disconnect them on demand.
        """
        sig = self._leap_signaler
        conntrack = self.connect_and_track
        auth_err = self._authentication_error

        # provider_setup signals
        conntrack(sig.prov_name_resolution, self._provider_setup_intermediate)
        conntrack(sig.prov_https_connection, self._provider_setup_intermediate)
        conntrack(sig.prov_download_provider_info, self._load_provider_config)

        # provider_bootstrap signals
        conntrack(sig.prov_download_ca_cert, self._provider_setup_intermediate)
        # XXX missing check_ca_fingerprint connection
        conntrack(sig.prov_check_api_certificate, self._provider_config_loaded)

        conntrack(sig.prov_problem_with_provider, self._login_problem_provider)
        conntrack(sig.prov_cancelled_setup, self._set_login_cancelled)

        # Login signals
        conntrack(sig.srp_auth_ok, self._authentication_finished)

        auth_error = lambda: auth_err(self.tr("Unknown error."))
        conntrack(sig.srp_auth_error, auth_error)

        auth_server_error = lambda: auth_err(self.tr(
            "There was a server problem with authentication."))
        conntrack(sig.srp_auth_server_error, auth_server_error)

        auth_connection_error = lambda: auth_err(self.tr(
            "Could not establish a connection."))
        conntrack(sig.srp_auth_connection_error, auth_connection_error)

        auth_bad_user_or_password = lambda: auth_err(self.tr(
            "Invalid username or password."))
        conntrack(sig.srp_auth_bad_user_or_password, auth_bad_user_or_password)

        # Logout signals
        sig.srp_logout_ok.connect(self._logout_ok)
        sig.srp_logout_error.connect(self._logout_error)
        # sig.srp_not_logged_in_error.connect(self._not_logged_in_error)



# -*- coding: utf-8 -*-
# mainwindow.py
# Copyright (C) 2013, 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Main window for Bitmask.
"""
import os
import time

import ConfigParser

from datetime import datetime

import psutil

from PySide import QtCore, QtGui

from leap.bitmask import __version__ as VERSION
from leap.bitmask import __version_hash__ as VERSION_HASH

# TODO: we should use a more granular signaling instead of passing error/ok as
# a result.
from leap.bitmask.backend.leapbackend import ERROR_KEY, PASSED_KEY

from leap.bitmask.config import flags
from leap.bitmask.logs.utils import get_logger, LOG_CONTROLLER

from leap.bitmask.gui.advanced_key_management import AdvancedKeyManagement
from leap.bitmask.gui.logwindow import LoggerWindow
from leap.bitmask.gui.login import LoginWidget
from leap.bitmask.gui.preferenceswindow import PreferencesWindow
from leap.bitmask.gui.signaltracker import SignalTracker
from leap.bitmask.gui.systray import SysTray
from leap.bitmask.gui.wizard import Wizard
from leap.bitmask.gui.providers import Providers
from leap.bitmask.gui.account import Account
from leap.bitmask.gui.app import App

from leap.bitmask.platform_init import IS_WIN, IS_MAC, IS_LINUX
from leap.bitmask.platform_init import locks
from leap.bitmask.platform_init.initializers import init_platform
from leap.bitmask.platform_init.initializers import init_signals

from leap.bitmask.util import autostart, make_address
from leap.bitmask.util.keyring_helpers import has_keyring

from leap.common.events import register
from leap.common.events import catalog

from leap.mail.imap.service.imap import IMAP_PORT

from ui_mainwindow import Ui_MainWindow

from leap.bitmask._components import HAS_EIP, HAS_MAIL

if HAS_EIP:
    from leap.bitmask.gui.eip_status import EIPStatusWidget
    from leap.bitmask.services.eip import conductor as eip_conductor
    from leap.bitmask.services import EIP_SERVICE

if HAS_MAIL:
    from leap.bitmask.gui.mail_status import MailStatusWidget
    from leap.bitmask.services.mail import conductor as mail_conductor
    from leap.bitmask.services import MX_SERVICE

QtDelayedCall = QtCore.QTimer.singleShot

logger = get_logger()

if not HAS_EIP:
    BITMASK_MAIL_ONLY_ICON = ":/images/menubar-mask-icon.png"


class MainWindow(QtGui.QMainWindow, SignalTracker):
    """
    Main window for login and presenting status updates to the user
    """
    # Signals

    new_updates = QtCore.Signal(object)
    raise_window = QtCore.Signal([])
    soledad_ready = QtCore.Signal([])
    all_services_stopped = QtCore.Signal()

    if HAS_EIP:
        eip_needs_login = QtCore.Signal([])
        # We use this flag to detect abnormal terminations
        user_stopped_eip = False

        # We give EIP some time to come up before starting soledad anyway
        EIP_START_TIMEOUT = 60000  # in milliseconds

    # We give the services some time to a halt before forcing quit.
    SERVICES_STOP_TIMEOUT = 3000  # in milliseconds

    def __init__(self, start_hidden=False, backend_pid=None):
        """
        Constructor for the client main window

        :param start_hidden: Set to true if the app should not show the window
                             but just the tray.
        :type start_hidden: bool
        """
        QtGui.QMainWindow.__init__(self)
        SignalTracker.__init__(self)

        autostart.set_autostart(True)

        # register leap events ########################################
        register(event=catalog.UPDATER_NEW_UPDATES,
                 callback=self._new_updates_available)  # make rpc call async
        register(event=catalog.RAISE_WINDOW,
                 callback=self._on_raise_window_event)  # make rpc call async
        # end register leap events ####################################

        self._updates_content = ""

        # setup UI
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.menuBar().setNativeMenuBar(not IS_LINUX)

        self.app = App()
        self._backend = self.app.backend
        self._leap_signaler = self.app.signaler
        self._settings = self.app.settings
        self._backend_settings = self._backend.settings

        # Login Widget
        self._login_widget = LoginWidget(self._backend,
                                         self._leap_signaler, self)
        self.ui.loginLayout.addWidget(self._login_widget)

        if HAS_MAIL:
            # Mail Widget
            self._mail_status = MailStatusWidget(self)
            self.ui.mailLayout.addWidget(self._mail_status)

        # Provider List
        self._providers = Providers(self.ui.cmbProviders)

        # Qt Signal Connections #####################################
        # TODO separate logic from ui signals.

        self.app.service_selection_changed.connect(
            self._update_eip_enabled_status)

        self._login_widget.login_finished.connect(
            self._on_user_logged_in)
        self._login_widget.logged_out.connect(self._on_user_logged_out)

        self._providers.connect_provider_changed(self._on_provider_changed)

        if HAS_EIP:
            # EIP Control redux #########################################
            self._eip_conductor = eip_conductor.EIPConductor(
                self._settings, self._backend, self._leap_signaler)
            self._eip_status = EIPStatusWidget(self, self._eip_conductor,
                                               self._leap_signaler)

            init_signals.eip_missing_helpers.connect(
                self._disable_eip_missing_helpers)

            self.ui.eipLayout.addWidget(self._eip_status)

            # XXX we should get rid of the circular refs
            # conductor <-> status,
            # right now keeping state on the widget ifself.
            self._eip_conductor.add_eip_widget(self._eip_status)

            self._eip_conductor.connect_signals()
            self._eip_conductor.qtsigs.connecting_signal.connect(
                self._on_eip_connecting)
            self._eip_conductor.qtsigs.connected_signal.connect(
                self._on_eip_connection_connected)
            self._eip_conductor.qtsigs.disconnected_signal.connect(
                self._on_eip_connection_disconnected)
            self._eip_conductor.qtsigs.connected_signal.connect(
                self._maybe_run_soledad_setup_checks)

            self.eip_needs_login.connect(self._eip_status.disable_eip_start)
            self.eip_needs_login.connect(self._disable_eip_start_action)

            # XXX all this info about state should move to eip conductor too
            self._already_started_eip = False
            self._trying_to_start_eip = False

        self._login_widget.login_offline_finished.connect(
            self._maybe_run_soledad_setup_checks)

        # This is created once we have a valid provider config
        self._logged_in_offline = False

        # Set used to track the services being stopped and need wait.
        self._services_being_stopped = {}

        # used to know if we are in the final steps of quitting
        self._quitting = False
        self._finally_quitting = False
        self._system_quit = False

        # Used to differentiate between a real quit and a close to tray
        self._close_to_tray = True

        self._backend_connect()

        self.ui.action_preferences.triggered.connect(self._show_preferences)
        self.ui.action_about_leap.triggered.connect(self._about)
        self.ui.action_quit.triggered.connect(self.quit)
        self.ui.action_wizard.triggered.connect(self._show_wizard)
        self.ui.action_show_logs.triggered.connect(self._show_logger_window)

        self.ui.action_help.setVisible(True)
        self.ui.action_help.triggered.connect(self._help)

        self.ui.action_create_new_account.triggered.connect(
            self._on_provider_changed)

        # Action item hidden since we don't provide stable mail yet.
        # TODO enable for 0.9.0 release??
        # self.ui.action_advanced_key_management.triggered.connect(
        #     self._show_AKM)

        if IS_MAC:
            self.ui.menuFile.menuAction().setText(self.tr("File"))

        self.raise_window.connect(self._do_raise_mainwindow)

        self._systray = None

        # XXX separate actions into a different module.
        if HAS_MAIL:
            self._action_mail_status = QtGui.QAction(
                self.tr("Mail is OFF"), self)
            self._mail_status.set_action_mail_status(
                self._action_mail_status)

        if HAS_EIP:
            self._action_eip_startstop = QtGui.QAction("", self)
            self._eip_status.set_action_eip_startstop(
                self._action_eip_startstop)

        self._action_visible = QtGui.QAction(self.tr("Show Main Window"), self)
        self._action_visible.triggered.connect(self._ensure_visible)

        self._enabled_services = []
        self._ui_mx_visible = True
        self._ui_eip_visible = True

        self._provider_details = None

        # last minute UI manipulations

        self._center_window()
        self.ui.lblNewUpdates.setVisible(False)
        self.ui.btnMore.setVisible(False)
        #########################################
        # We hide this in height temporarily too
        self.ui.lblNewUpdates.resize(0, 0)
        self.ui.btnMore.resize(0, 0)
        #########################################
        self.ui.btnMore.clicked.connect(self._updates_details)

        if flags.OFFLINE:
            self._login_widget._set_label_offline()

        # Services signals/slots connection
        self.new_updates.connect(self._react_to_new_updates)

        self.soledad_ready.connect(self._start_mail_service)
        # ################################ end Qt Signals connection ########

        self._wizard = None
        self._wizard_firstrun = False

        self._start_hidden = start_hidden
        self._backend_pid = backend_pid

        if HAS_MAIL:
            self._mail_conductor = mail_conductor.MailConductor(self._backend)
            self._mail_conductor.connect_mail_signals(self._mail_status)

        if not init_platform():
            self.quit()
            return

        # start event machines from within the eip and mail conductors
        # TODO should encapsulate all actions into one object
        if HAS_EIP:
            self._eip_conductor.start_eip_machine(
                action=self._action_eip_startstop)
        if HAS_MAIL:
            self._mail_conductor.start_mail_machine()

        if self._first_run():
            self._wizard_firstrun = True
            self._show_wizard()
        else:
            # during finish_init, we disable the eip start button
            # so this has to be done after eip_machine is started
            self._finish_init()

    def _not_logged_in_error(self):
        """
        Handle the 'not logged in' backend error if we try to do an operation
        that requires to be logged in.
        """
        logger.critical("You are trying to do an operation that requires "
                        "log in first.")
        QtGui.QMessageBox.critical(
            self, self.tr("Application error"),
            self.tr("You are trying to do an operation "
                    "that requires logging in first."))

    def _backend_bad_call(self, data):
        """
        Callback for debugging bad backend calls

        :param data: data from the backend about the problem
        :type data: str
        """
        logger.error("Bad call to the backend:")
        logger.error(data)

    def _backend_connect(self, only_tracked=False):
        """
        Connect to backend signals.

        We track some signals in order to disconnect them on demand.
        For instance, in the wizard we need to connect to some signals that are
        already connected in the mainwindow, so to avoid conflicts we do:
            - disconnect signals needed in wizard (`disconnect_and_untrack`)
            - use wizard
            - reconnect disconnected signals (we use the `only_tracked` param)

        :param only_tracked: whether or not we should connect only the signals
                             that we are tracking to disconnect later.
        :type only_tracked: bool
        """
        sig = self._leap_signaler
        conntrack = self.connect_and_track
        # XXX does this goes in here? this will be triggered when the login or
        # wizard requests provider data

        conntrack(sig.prov_check_api_certificate, self._get_provider_details)
        conntrack(sig.prov_get_details, self._provider_get_details)

        # EIP bootstrap signals
        conntrack(sig.eip_config_ready, self._eip_intermediate_stage)
        conntrack(sig.eip_client_certificate_ready, self._finish_eip_bootstrap)

        ###################################################
        # Add tracked signals above this, untracked below!
        ###################################################
        if only_tracked:
            return

        # We don't want to disconnect some signals so don't track them:

        sig.backend_bad_call.connect(self._backend_bad_call)

        sig.prov_unsupported_client.connect(self._needs_update)
        sig.prov_unsupported_api.connect(self._incompatible_api)
        sig.prov_get_all_services.connect(self._provider_get_all_services)

        # Logout signals =================================================
        # This error may be due a 'logout' or a 'password change', as is used
        # on the login widget and the settings dialog the connection is made
        # here.
        sig.srp_not_logged_in_error.connect(self._not_logged_in_error)

        if HAS_EIP:
            # EIP start signals ==============================================
            self._eip_conductor.connect_backend_signals()
            sig.eip_can_start.connect(self._backend_can_start_eip)
            sig.eip_cannot_start.connect(self._backend_cannot_start_eip)

            sig.eip_dns_error.connect(self._eip_dns_error)

            sig.eip_get_gateway_country_code.connect(self._set_eip_provider)
            sig.eip_no_gateway.connect(self._set_eip_provider)
            # ==================================================================

        # Soledad signals
        # TODO delegate connection to soledad bootstrapper
        sig.soledad_bootstrap_failed.connect(
            self._mail_status.set_soledad_failed)
        sig.soledad_bootstrap_finished.connect(self._on_soledad_ready)

        sig.soledad_offline_failed.connect(
            self._mail_status.set_soledad_failed)
        sig.soledad_offline_finished.connect(self._on_soledad_ready)

        sig.soledad_invalid_auth_token.connect(
            self._mail_status.set_soledad_invalid_auth_token)

        # TODO: connect this with something
        # sig.soledad_cancelled_bootstrap.connect()

    def _show_wizard(self):
        """
        TRIGGERS:
            self.ui.action_wizard.triggered

        This is called during the Bitmask's first run.

        Open the wizard window, hiding the mainwindow window.
        This reuses an existing wizard object or creates a new one if needed.
        """
        # Disconnect signals to avoid conflicts between mainwindow and wizard
        # connections
        self.disconnect_and_untrack()

        if self._wizard is None:
            self._wizard = Wizard(backend=self._backend,
                                  leap_signaler=self._leap_signaler)
            self._wizard.accepted.connect(self._wizard_accepted)
            self._wizard.rejected.connect(self._wizard_rejected)
            self._wizard.finished.connect(self._wizard_finished)

        self.setVisible(False)
        # Do NOT use exec_, it will use a child event loop!
        # Refer to http://www.themacaque.com/?p=1067 for funny details.
        self._wizard.show()
        if IS_MAC:
            self._wizard.raise_()
        self._settings.set_skip_first_run(True)

    def _wizard_finished(self):
        """
        TRIGGERS:
            self._wizard.finished

        Called when the wizard has finished.
        """
        self.setVisible(True)
        self._backend_connect(only_tracked=True)

    def _wizard_rejected(self):
        """
        TRIGGERS:
            self._wizard.rejected

        Called if the wizard has been cancelled or closed before finishing.
        """
        # NOTE: We need to verify if we have configured providers.
        # If not then we should quit
        # see self._rejected_wizard()
        providers = self._settings.get_configured_providers()
        has_provider_on_disk = len(providers) != 0
        if not has_provider_on_disk:
            # if we don't have any provider configured (included a pinned
            # one) we can't use the application, so quit.
            self._wizard = None
            self.quit()
        else:
            # This happens if the user finishes the provider
            # setup but does not register
            self._wizard = None

            # self._backend_connect(only_tracked=True)
            # if self._wizard_firstrun:
            #     self._finish_init()

    def _wizard_accepted(self):
        """
        TRIGGERS:
            self._wizard.accepted

        Called when the wizard has accepted.
        """
        self.setVisible(True)
        self._finish_init()

    def _load_from_wizard(self):
        """
        Load the resulting information of the user going through the Wizard.
        Trigger the login sequence if needed.
        """
        possible_username = self._wizard.get_username()
        possible_password = self._wizard.get_password()

        # select the configured provider in the combo box
        domain = self._wizard.get_domain()
        self._providers.select_provider_by_name(domain)
        self._login_widget.set_provider(domain)

        self._login_widget.set_remember(self._wizard.get_remember())
        self._enabled_services = list(self._wizard.get_services())
        self._settings.set_enabled_services(
            self._providers.get_selected_provider(),
            self._enabled_services)

        if possible_username is not None:
            self._login_widget.set_user(possible_username)
        if possible_password is not None:
            self._login_widget.set_password(possible_password)
            self._login()
        else:
            if HAS_EIP:
                self.eip_needs_login.emit()

        self._wizard = None

    def _show_logger_window(self):
        """
        TRIGGERS:
            self.ui.action_show_logs.triggered

        Display the window with the history of messages logged until now
        and displays the new ones on arrival.
        """
        lw = LoggerWindow(self)
        lw.show()

    def _show_AKM(self):
        """
        TRIGGERS:
            self.ui.action_advanced_key_management.triggered

        Display the Advanced Key Management dialog.
        """
        logged_user = self._login_widget.get_logged_user()

        details = self._provider_details
        mx_provided = False
        if details is not None:
            mx_provided = MX_SERVICE in details['services']

        # XXX: handle differently not logged in user?
        akm = AdvancedKeyManagement(self, mx_provided, logged_user,
                                    self._backend, self.app.soledad_started)
        akm.show()

    def _show_preferences(self):
        """
        TRIGGERS:
            self.ui.btnPreferences.clicked (disabled for now)
            self.ui.action_preferences

        Display the preferences window.
        """
        logged_user = self._login_widget.get_logged_user()
        if logged_user is not None:
            user, domain = logged_user.split('@')
        else:
            user = None
            domain = self._providers.get_selected_provider()

        account = Account(user, domain)
        pref_win = PreferencesWindow(self, account, self.app)
        pref_win.show()

    def _update_eip_enabled_status(self, account=None, services=None):
        """
        TRIGGER:
            App.service_selection_changed

        Enable or disable the EIP start/stop actions and stop EIP if the user
        disabled that service.

        :returns: if the eip actions were enabled or disabled
        :rtype: bool
        """
        if account is not None:
            domain = account.domain
        else:
            # I am not sure why, but asking for the currently selected
            # provider here give you the WRONG provider
            domain = self.app.settings.get_defaultprovider()

        if domain is None:
            logger.warning("Trying to update eip enabled status but there's no"
                           " default provider. Disabling EIP for the time"
                           " being...")
            self._backend_cannot_start_eip()
            return

        services_enabled = self.app.settings.get_enabled_services(domain)

        if HAS_EIP:
            if EIP_SERVICE not in services_enabled:
                self._eip_conductor.terminate()

                def hide():
                    self.app.backend.eip_can_start(domain=domain)

                QtDelayedCall(100, hide)
                # ^^ VERY VERY Hacky, but with the simple state machine,
                # there is no way to signal 'disconnect and then disable'

            else:
                settings = self.app.settings
                self._trying_to_start_eip = settings.get_autostart_eip()
                if not self._trying_to_start_eip:
                    self._backend.eip_setup(provider=domain, skip_network=True)
                # check if EIP can start (will trigger widget update)
                self.app.backend.eip_can_start(domain=domain)

    def _backend_can_start_eip(self):
        """
        TRIGGER:
            self._backend.signaler.eip_can_start

        If EIP can be started right away, and the client is configured
        to do so, start it. Otherwise it leaves everything in place
        for the user to click Turn ON.
        """
        if self._eip_status.missing_helpers:
            self._eip_status.disable_eip_start()
            return

        settings = self._settings
        default_provider = settings.get_defaultprovider()
        enabled_services = []
        if default_provider is not None:
            enabled_services = settings.get_enabled_services(default_provider)
        eip_enabled = False
        if EIP_SERVICE in enabled_services:
            eip_enabled = True
            if default_provider is not None:
                self._eip_status.enable_eip_start()
                self._eip_status.set_eip_status("")
            else:
                # we don't have an usable provider
                # so the user needs to log in first
                self._eip_status.disable_eip_start()
        else:
            self._eip_status.disable_eip_start()
            self._eip_status.set_eip_status(self.tr("Disabled"))

        if eip_enabled and self._trying_to_start_eip:
            self._trying_to_start_eip = False
            self._try_autostart_eip()

    def _backend_cannot_start_eip(self):
        """
        TRIGGER:
            self._backend.signaler.eip_cannot_start

        If EIP can't be started right away, get the UI to what it
        needs to look like and waits for a proper login/eip bootstrap.
        """
        settings = self._settings
        default_provider = settings.get_defaultprovider()
        enabled_services = []
        if default_provider is not None:
            enabled_services = settings.get_enabled_services(default_provider)

        if HAS_EIP:
            if EIP_SERVICE in enabled_services:
                # we don't have a usable provider
                # so the user needs to log in first
                self._eip_status.disable_eip_start()
            else:
                self._eip_status.disable_eip_start()
                # NOTE: we shouldn't be setting the message here.
                if not self._eip_status.missing_helpers:
                    self._eip_status.set_eip_status(self.tr("Disabled"))

        # this state flag is responsible for deferring the login
        # so we must update it, otherwise we're in a deadlock.
        self._trying_to_start_eip = False

    def _disable_eip_missing_helpers(self):
        """
        TRIGGERS:
            init_signals.missing_helpers

        Set the missing_helpers flag, so we can disable EIP.
        """
        self._eip_status.missing_helpers = True

    #
    # updates
    #

    def _new_updates_available(self, event, content):
        """
        Callback for the new updates event

        :param event: The event that triggered the callback.
        :type event: str
        :param content: The content of the event.
        :type content: list
        """
        self.new_updates.emit(content)

    def _react_to_new_updates(self, content):
        """
        TRIGGERS:
            self.new_updates

        Display the new updates label and sets the updates_content

        :param content: The content of the event.
        :type content: list
        """
        self.moveToThread(QtCore.QCoreApplication.instance().thread())
        self.ui.lblNewUpdates.setVisible(True)
        self.ui.btnMore.setVisible(True)
        self._updates_content = content

    def _updates_details(self):
        """
        TRIGGERS:
            self.ui.btnMore.clicked

        Parses and displays the updates details
        """
        msg = self.tr("The Bitmask app is ready to update, please"
                      " restart the application.")

        # We assume that if there is nothing in the contents, then
        # the Bitmask bundle is what needs updating.
        if len(self._updates_content) > 0:
            files = self._updates_content.split(", ")
            files_str = ""
            for f in files:
                final_name = f.replace("/data/", "")
                final_name = final_name.replace(".thp", "")
                files_str += final_name
                files_str += "\n"
            msg += self.tr(" The following components will be updated:\n%s") \
                % (files_str,)

        QtGui.QMessageBox.information(self,
                                      self.tr("Updates available"),
                                      msg)

    def _finish_init(self):
        """
        TRIGGERS:
            self._wizard.accepted

        Also called at the end of the constructor if not first run.

        Implements the behavior after either constructing the
        mainwindow object, loading the saved user/password, or after
        the wizard has been executed.
        """
        # XXX: May be this can be divided into two methods?
        providers = self._settings.get_configured_providers()
        self._providers.set_providers(providers)

        provider = self._providers.get_selected_provider()
        self._login_widget.set_provider(provider)

        self._show_systray()

        if not self._start_hidden:
            self.show()
            if IS_MAC:
                self.raise_()

        self._show_hide_unsupported_services()

        LOG_CONTROLLER.start_logbook_subscriber()

        # XXX - HACK, kind of...
        # With the 1ms QTimer.singleShot call we schedule the call right after
        # other signals waiting for the qt reactor to take control.
        # That way, the method is called right after the EIP machines' signals.
        # We need to wait until that happens because the state-machine
        # controlled widget shows the 'Turn On' button and we want to do the
        # changes to that button right after, not before.
        QtDelayedCall(1, self._update_eip_enabled_status)

        if self._wizard:
            self._load_from_wizard()
        else:
            if self._load_credentials_from_env():
                self._login()
                return

            domain = self._settings.get_provider()
            if domain is not None:
                self._providers.select_provider_by_name(domain)
                self._login_widget.set_provider(domain)

            if not self._settings.get_remember():
                # nothing to do here
                return

            saved_user = self._settings.get_user()

            if saved_user is not None and has_keyring():
                if self._login_widget.load_user_from_keyring(saved_user):
                    self._login()

    def _load_credentials_from_env(self):
        """
        Load username and password into the login widget from a file specified
        in an environment variable. This is useful for test purposes.

        :return: True if credentials were loaded, False otherwise
        :rtype: bool
        """
        credentials = os.environ.get("BITMASK_CREDENTIALS")

        if credentials is None:
            return False

        try:
            config = ConfigParser.ConfigParser()
            config.read(credentials)
            username, domain = config.get('Credentials', 'username').split('@')
            password = config.get('Credentials', 'password')
        except Exception, e:
            print "Error reading credentials file: {0}".format(e)
            return False

        self._providers.select_provider_by_name(domain)
        self._login_widget.set_provider(domain)
        self._login_widget.set_user(username)
        self._login_widget.set_password(password)

        return True

    def _show_hide_unsupported_services(self):
        """
        Given a set of configured providers, it creates a set of
        available services among all of them and displays the service
        widgets of only those.

        This means, for example, that with just one provider with EIP
        only, the mail widget won't be displayed.
        """
        providers = self._settings.get_configured_providers()
        self._backend.provider_get_all_services(providers=providers)

    def _provider_get_all_services(self, services):
        if HAS_EIP:
            self._set_eip_visible(EIP_SERVICE in services)
        if HAS_MAIL:
            self._set_mx_visible(MX_SERVICE in services)

    def _set_mx_visible(self, visible):
        """
        Change the visibility of MX_SERVICE related UI components.

        :param visible: whether the components should be visible or not.
        :type visible: bool
        """
        # only update visibility if it is something to change
        if self._ui_mx_visible ^ visible:
            self.ui.mailWidget.setVisible(visible)
            self.ui.lineUnderEmail.setVisible(visible)
            self._action_mail_status.setVisible(visible)
            self._ui_mx_visible = visible

    def _set_eip_visible(self, visible):
        """
        Change the visibility of EIP_SERVICE related UI components.

        :param visible: whether the components should be visible or not.
        :type visible: bool
        """
        # NOTE: we use xor to avoid the code being run if the visibility hasn't
        # changed. This is meant to avoid the eip menu being displayed floating
        # around at start because the systray isn't rendered yet.
        if self._ui_eip_visible ^ visible:
            self.ui.eipWidget.setVisible(visible)
            self.ui.lineUnderEIP.setVisible(visible)
            self._eip_menu.setVisible(visible)
            self._ui_eip_visible = visible

    #
    # systray
    #

    def _show_systray(self):
        """
        Sets up the systray icon
        """
        if self._systray is not None:
            self._systray.setVisible(True)
            return

        # Placeholder action
        # It is temporary to display the tray as designed
        help_action = QtGui.QAction(self.tr("Help"), self)
        help_action.setEnabled(False)

        systrayMenu = QtGui.QMenu(self)
        systrayMenu.addAction(self._action_visible)
        systrayMenu.addSeparator()

        if HAS_EIP:
            eip_status_label = u"{0}: {1}".format(
                self._eip_conductor.eip_name, self.tr("OFF"))
            self._eip_menu = eip_menu = systrayMenu.addMenu(eip_status_label)
            eip_menu.addAction(self._action_eip_startstop)
            self._eip_status.set_eip_status_menu(eip_menu)
            systrayMenu.addSeparator()
        if HAS_MAIL:
            systrayMenu.addAction(self._action_mail_status)
            systrayMenu.addSeparator()
        systrayMenu.addAction(self.ui.action_quit)

        self._systray = SysTray(self)
        self._systray.setContextMenu(systrayMenu)

        if HAS_EIP:
            self._systray.setIcon(self._eip_status.ERROR_ICON_TRAY)
        else:
            mail_status_icon = QtGui.QPixmap(BITMASK_MAIL_ONLY_ICON)
            self._systray.setIcon(mail_status_icon)

        self._systray.setVisible(True)
        self._systray.activated.connect(self._tray_activated)

        if HAS_EIP:
            self._eip_status.set_systray(self._systray)
        if HAS_MAIL:
            self._mail_status.set_systray(self._systray)

        if self._start_hidden:
            hello = lambda: self._systray.showMessage(
                self.tr('Hello!'),
                self.tr('Bitmask has started in the tray.'))
            # we wait for the systray to be ready
            QtDelayedCall(1000, hello)

    def _tray_activated(self, reason=None):
        """
        TRIGGERS:
            self._systray.activated

        :param reason: the reason why the tray got activated.
        :type reason: int

        Display the context menu from the tray icon
        """
        context_menu = self._systray.contextMenu()
        if not IS_MAC:
            # for some reason, context_menu.show()
            # is failing in a way beyond my understanding.
            # (not working the first time it's clicked).
            # this works however.
            context_menu.exec_(self._systray.geometry().center())

    def _ensure_visible(self):
        """
        TRIGGERS:
            self._action_visible.triggered

        Ensure that the window is visible and raised.
        """
        QtGui.QApplication.setQuitOnLastWindowClosed(True)
        self.show()
        if IS_LINUX:
            # On ubuntu, activateWindow doesn't work reliably, so
            # we do the following as a workaround. See
            # https://bugreports.qt-project.org/browse/QTBUG-24932
            # for more details
            QtGui.QX11Info.setAppUserTime(0)
        self.activateWindow()
        self.raise_()

    def _ensure_invisible(self):
        """
        TRIGGERS:
            self._action_visible.triggered

        Ensure that the window is hidden.
        """
        # We set this in order to avoid dialogs shutting down the
        # app on close, as they will be the only visible window.
        # e.g.: PreferencesWindow, LoggerWindow
        QtGui.QApplication.setQuitOnLastWindowClosed(False)
        self.hide()

    def _center_window(self):
        """
        Center the main window based on the desktop geometry
        """
        geometry = self._settings.get_geometry()
        state = self._settings.get_windowstate()

        if geometry is None:
            app = QtGui.QApplication.instance()
            width = app.desktop().width()
            height = app.desktop().height()
            window_width = self.size().width()
            window_height = self.size().height()
            x = (width / 2.0) - (window_width / 2.0)
            y = (height / 2.0) - (window_height / 2.0)
            self.move(x, y)
        else:
            self.restoreGeometry(geometry)

        if state is not None:
            self.restoreState(state)

    def _about(self):
        """
        TRIGGERS:
            self.ui.action_about_leap.triggered

        Display the About Bitmask dialog
        """
        today = datetime.now().date()
        greet = ("Happy New 1984!... or not ;)<br><br>"
                 if today.month == 1 and today.day < 15 else "")
        title = self.tr("About Bitmask - %s") % (VERSION,)
        msg = self.tr(
            "Version: <b>{ver}</b> ({ver_hash})<br>"
            "<br>{greet}"
            "Bitmask is the Desktop client application for the LEAP "
            "platform, supporting Encrypted Internet Proxy "
            "and <a href='https://bitmask.net/help/email'> "
            "Encrypted Email</a>.<br><br>"
            "LEAP is a non-profit dedicated to giving all internet users "
            "access to secure communication. Our focus is on adapting "
            "encryption technology to make it easy to use and widely "
            "available.<br>"
            "<br>"
            "<a href='https://leap.se'>More about LEAP</a>")
        msg = msg.format(ver=VERSION, ver_hash=VERSION_HASH[:10], greet=greet)
        QtGui.QMessageBox.about(self, title, msg)

    def _help(self):
        """
        TRIGGERS:
            self.ui.action_help.triggered

        Display the Bitmask help dialog.
        """
        # TODO: don't hardcode!
        smtp_port = 2013

        help_url = "<p><a href='https://{0}'>{0}</a></p>".format(
            self.tr("bitmask.net/help"))

        lang = QtCore.QLocale.system().name().replace('_', '-')
        thunderbird_extension_url = \
            "https://addons.mozilla.org/{0}/" \
            "thunderbird/addon/bitmask/".format(lang)

        email_quick_reference = self.tr("Email quick reference")
        thunderbird_text = self.tr(
            "For Thunderbird, you can use the "
            "Bitmask extension. Search for \"Bitmask\" in the add-on "
            "manager or download it from <a href='{0}'>"
            "addons.mozilla.org</a>.".format(thunderbird_extension_url))
        manual_text = self.tr(
            "Alternatively, you can manually configure "
            "your mail client to use Bitmask Email with these options:")
        manual_imap = self.tr("IMAP: localhost, port {0}".format(IMAP_PORT))
        manual_smtp = self.tr("SMTP: localhost, port {0}".format(smtp_port))
        manual_username = self.tr("Username: your full email address")
        manual_password = self.tr("Password: any non-empty text")

        msg = help_url + self.tr(
            "<p><strong>{0}</strong></p>"
            "<p>{1}</p>"
            "<p>{2}"
            "<ul>"
            "<li>&nbsp;{3}</li>"
            "<li>&nbsp;{4}</li>"
            "<li>&nbsp;{5}</li>"
            "<li>&nbsp;{6}</li>"
            "</ul></p>").format(email_quick_reference, thunderbird_text,
                                manual_text, manual_imap, manual_smtp,
                                manual_username, manual_password)
        QtGui.QMessageBox.about(self, self.tr("Bitmask Help"), msg)

    def _needs_update(self):
        """
        Display a warning dialog to inform the user that the app needs update.
        """
        url = "https://dl.bitmask.net/"
        msg = self.tr(
            "The current client version is not supported "
            "by this provider.<br>"
            "Please update to latest version.<br><br>"
            "You can get the latest version from "
            "<a href='{0}'>{1}</a>").format(url, url)
        QtGui.QMessageBox.warning(self, self.tr("Update Needed"), msg)

    def _incompatible_api(self):
        """
        Display a warning dialog to inform the user that the provider has an
        incompatible API.
        """
        msg = self.tr(
            "This provider is not compatible with the client.<br><br>"
            "Error: API version incompatible.")
        QtGui.QMessageBox.warning(self, self.tr("Incompatible Provider"), msg)

    def closeEvent(self, e):
        """
        Reimplementation of closeEvent to close to tray
        """
        if not e.spontaneous():
            # if the system requested the `close` then we should quit.
            self._system_quit = True
            self.quit(disable_autostart=False)
            return

        if QtGui.QSystemTrayIcon.isSystemTrayAvailable() and \
                self._close_to_tray:
            self._ensure_invisible()
            e.ignore()
            return

        self._settings.set_geometry(self.saveGeometry())
        self._settings.set_windowstate(self.saveState())

        QtGui.QMainWindow.closeEvent(self, e)

    def _first_run(self):
        """
        Return True if there are no configured providers. False otherwise

        :rtype: bool
        """
        providers = self._settings.get_configured_providers()
        has_provider_on_disk = len(providers) != 0
        skip_first_run = self._settings.get_skip_first_run()
        return not (has_provider_on_disk and skip_first_run)

    def _disconnect_login_wait(self):
        """
        Disconnect the EIP finishing signal to the wait flag on the login
        widget.
        """
        try:
            eip_sigs = self._eip_conductor.qtsigs
            slot = lambda: self._login_widget.wait_for_login(False)

            eip_sigs.connected_signal.disconnect(slot)
            eip_sigs.disconnected_signal.disconnect(slot)
            eip_sigs.connection_aborted_signal.disconnect(slot)
            eip_sigs.connection_died_signal.disconnect(slot)
        except Exception:
            # signal not connected
            pass

    def _on_provider_changed(self, wizard=True):
        """
        TRIGGERS:
            self._providers._provider_changed
            self.ui.action_create_new_account.triggered

        Ask the user if really wants to change provider since a services stop
        is required for that action.

        :param wizard: whether the 'other...' option was picked or not.
        :type wizard: bool
        """
        # TODO: we should handle the case that EIP is autostarting since we
        # won't get a warning until EIP has fully started.
        # TODO: we need to add a check for the mail status (smtp/imap/soledad)

        something_runing = self._login_widget.get_logged_user() is not None
        if HAS_EIP:
            something_runing = something_runing or self._already_started_eip

        provider = self._providers.get_selected_provider()
        self._login_widget.set_provider(provider)

        if not something_runing:
            if wizard:
                self._show_wizard()
            else:
                self._settings.set_provider(provider)
                self._settings.set_defaultprovider(provider)
                self._update_eip_enabled_status()
            return

        title = self.tr("Stop services")
        text = "<b>" + self.tr("Do you want to stop all services?") + "</b>"
        informative_text = self.tr("In order to change the provider, the "
                                   "running services needs to be stopped.")

        msg = QtGui.QMessageBox(self)
        msg.setWindowTitle(title)
        msg.setText(text)
        msg.setInformativeText(informative_text)
        msg.setStandardButtons(QtGui.QMessageBox.Yes | QtGui.QMessageBox.No)
        msg.setDefaultButton(QtGui.QMessageBox.No)
        msg.setIcon(QtGui.QMessageBox.Warning)
        res = msg.exec_()

        if res == QtGui.QMessageBox.Yes:
            self._settings.set_provider(provider)
            self._settings.set_defaultprovider(provider)
            self._settings.set_autostart_eip(False)
            self._stop_services()
            self._update_eip_enabled_status()
            self._eip_conductor.qtsigs.do_disconnect_signal.emit()
            if wizard:
                self._show_wizard()
        else:
            if not wizard:
                # if wizard, the widget restores itself
                self._providers.restore_previous_provider()

    def _login(self):
        """
        TRIGGERS:
            self._login_widget.login

        Start the login sequence. Which involves bootstrapping the
        selected provider if the selection is valid (not empty), then
        start the SRP authentication, and as the last step
        bootstrapping the EIP service
        """
        self.ui.action_create_new_account.setEnabled(False)

        ok = self._login_widget.do_login()
        if not ok:
            logger.error("There was a problem triggering the login.")
            return

    def _cancel_ongoing_defers(self):
        """
        Cancel the running defers to avoid app blocking.
        """
        # XXX: Should we stop all the backend defers?
        self._backend.soledad_cancel_bootstrap()
        self._backend.soledad_close()

        self.app.soledad_started = True

    def _on_user_logged_in(self):
        """
        TRIGGERS:
            self._login_widget.logged_in

        Once the user is properly authenticated, try starting the EIP
        service.
        """
        self._disconnect_login_wait()

        user = self._login_widget.get_logged_user()
        # XXX the widget now gives us the full user id.
        # this is confusing.

        self.ui.action_create_new_account.setEnabled(True)

        if HAS_EIP:
            self._start_eip_bootstrap()
        if HAS_MAIL:
            # XXX the casting to str (needed by smtp gateway) should be done
            # in a better place.
            self._mail_conductor.userid = str(user)
            if MX_SERVICE in self._enabled_services:
                btn_enabled = self._login_widget.set_logout_btn_enabled
                btn_enabled(False)
                sig = self._leap_signaler
                sig.soledad_bootstrap_failed.connect(
                    lambda: btn_enabled(True))
                sig.soledad_bootstrap_finished.connect(
                    lambda: btn_enabled(True))

            if MX_SERVICE not in self._provider_details['services']:
                self._set_mx_visible(False)

            if not HAS_EIP:
                # This has to be worked out in Bitmask 0.10.
                # Since EIP won't start, we need to trigger
                # the soledad setup service from here.
                self._maybe_run_soledad_setup_checks()

    def _on_user_logged_out(self):
        """
        TRIGGER:
            self._login_widget.logged_out

        Switch the stackedWidget back to the login stage after
        logging out
        """
        if HAS_MAIL:
            self._mail_conductor.stop_mail_services()
            self._mail_status.mail_state_disabled()
        self._show_hide_unsupported_services()

    def _start_eip_bootstrap(self):
        """
        Change the stackedWidget index to the EIP status one and
        triggers the eip bootstrapping.
        """
        domain = self._providers.get_selected_provider()
        self._enabled_services = self._settings.get_enabled_services(domain)

        # TODO separate UI from logic.
        if self._provides_mx_and_enabled():
            self._mail_status.about_to_start()
        else:
            self._mail_status.set_disabled()

        self._maybe_start_eip()

    def _get_provider_details(self, data):
        """
        TRIGGERS:
            prov_check_api_certificate

        Set the attributes to know if the EIP and MX services are supported
        and enabled.
        This is triggered right after the provider has been set up.
        """
        domain = self._providers.get_selected_provider()
        lang = QtCore.QLocale.system().name()
        self._backend.provider_get_details(domain=domain, lang=lang)

    def _provider_get_details(self, details):
        """
        Set the details for the just downloaded provider.

        :param details: the details of the provider.
        :type details: dict
        """
        self._provider_details = details

    def _provides_mx_and_enabled(self):
        """
        Define if the current provider provides mx and if we have it enabled.

        :returns: True if provides and is enabled, False otherwise
        :rtype: bool
        """
        domain = self._providers.get_selected_provider()
        enabled_services = self._settings.get_enabled_services(domain)

        mx_enabled = MX_SERVICE in enabled_services
        mx_provided = False
        if self._provider_details is not None:
            mx_provided = MX_SERVICE in self._provider_details['services']

        return mx_enabled and mx_provided

    def _provides_eip_and_enabled(self):
        """
        Define if the current provider provides eip and if we have it enabled.

        :returns: True if provides and is enabled, False otherwise
        :rtype: bool
        """
        domain = self._providers.get_selected_provider()
        enabled_services = self._settings.get_enabled_services(domain)

        eip_enabled = EIP_SERVICE in enabled_services
        eip_provided = False
        if self._provider_details is not None:
            eip_provided = EIP_SERVICE in self._provider_details['services']

        return eip_enabled and eip_provided

    def _maybe_run_soledad_setup_checks(self):
        """
        TRIGGERS:
            self._eip_conductor.qtsigs.connected_signal
            self._login_widget.login_offline_finished

        Conditionally start Soledad.
        """
        # TODO split.
        if not self._provides_mx_and_enabled() and not flags.OFFLINE:
            logger.debug("Provider does not offer MX, but it is enabled.")
            return

        username = self._login_widget.get_user()
        password = unicode(self._login_widget.get_password())
        provider_domain = self._providers.get_selected_provider()

        if flags.OFFLINE:
            full_user_id = make_address(username, provider_domain)
            uuid = self._backend_settings.get_uuid(full_user_id)
            self._mail_conductor.userid = full_user_id

            if not uuid:
                # We don't need more visibility at the moment,
                # this is mostly for internal use/debug for now.
                logger.warning("Sorry! Log-in at least one time.")
                return
            self._backend.soledad_load_offline(username=full_user_id,
                                               password=password, uuid=uuid)
        else:
            logged_user = self._login_widget.get_logged_user()
            if logged_user is not None:
                username, domain = logged_user.split('@')
                self._backend.soledad_bootstrap(username=username,
                                                domain=domain,
                                                password=password)

    ###################################################################
    # Service control methods: soledad

    def _on_soledad_ready(self):
        """
        TRIGGERS:
            Signaler.soledad_bootstrap_finished

        Actions to take when Soledad is ready.
        """
        logger.debug("Done bootstrapping Soledad")

        self.app.soledad_started = True
        self.soledad_ready.emit()

    ###################################################################
    # Service control methods: mail
    def _start_mail_service(self):
        if self._provides_mx_and_enabled() or flags.OFFLINE:
            self._mail_conductor.start_mail_service(download_if_needed=True,
                                                    offline=flags.OFFLINE)

    ###################################################################
    # Service control methods: eip

    def _disable_eip_start_action(self):
        """
        Disable the EIP start action in the systray menu.
        """
        self._action_eip_startstop.setEnabled(False)

    def _enable_eip_start_action(self):
        """
        Enable the EIP start action in the systray menu.
        """
        self._action_eip_startstop.setEnabled(True)
        self._eip_status.enable_eip_start()

    def _on_eip_connecting(self):
        """
        TRIGGERS:
            self._eip_conductor.qtsigs.connecting_signal

        This is triggered when EIP starts connecting.

        React to any EIP finishing signal[1] that means that the network is
        ready to be used and trigger the "don't keep waiting" action on the
        login widget.

        [1] finishing signal => connected, disconnected, aborted or died
        """
        self._login_widget.wait_for_login(True)

        eip_sigs = self._eip_conductor.qtsigs
        slot = lambda: self._login_widget.wait_for_login(False)

        eip_sigs.connected_signal.connect(slot)
        eip_sigs.disconnected_signal.connect(slot)
        eip_sigs.connection_aborted_signal.connect(slot)
        eip_sigs.connection_died_signal.connect(slot)

    def _on_eip_connection_connected(self):
        """
        TRIGGERS:
            self._eip_conductor.qtsigs.connected_signal

        This is a little workaround for connecting the vpn-connected
        signal that currently is beeing processed under status_panel.
        After the refactor to EIPConductor this should not be necessary.
        """
        self._already_started_eip = True

        domain = self._providers.get_selected_provider()
        self._settings.set_defaultprovider(domain)

        self._backend.eip_get_gateway_country_code(domain=domain)

        # check for connectivity
        self._backend.eip_check_dns(domain=domain)

    def _on_eip_connection_disconnected(self):
        """
        TRIGGERS:
            self._eip_conductor.qtsigs.disconnected_signal

        Set the eip status to not started.
        """
        self._already_started_eip = False

    def _set_eip_provider(self, country_code=None):
        """
        TRIGGERS:
            Signaler.eip_get_gateway_country_code
            Signaler.eip_no_gateway

        Set the current provider and country code in the eip status widget.
        """
        domain = self._providers.get_selected_provider()
        self._eip_status.set_provider(domain, country_code)

    def _eip_dns_error(self):
        """
        Trigger this if we don't have a working DNS resolver.
        """
        domain = self._providers.get_selected_provider()
        msg = self.tr(
            "The server at {0} can't be found, because the DNS lookup "
            "failed. DNS is the network service that translates a "
            "website's name to its Internet address. Either your computer "
            "is having trouble connecting to the network, or you are "
            "missing some helper files that are needed to securely use "
            "DNS while {1} is active. To install these helper files, quit "
            "this application and start it again."
        ).format(domain, self._eip_conductor.eip_name)

        QtGui.QMessageBox.critical(self, self.tr("Connection Error"), msg)

    def _try_autostart_eip(self):
        """
        Try to autostart EIP.
        """
        settings = self._settings
        default_provider = settings.get_defaultprovider()
        self._enabled_services = settings.get_enabled_services(
            default_provider)

        if settings.get_autostart_eip():
            self._maybe_start_eip(autostart=True)

    # eip boostrapping, config etc...

    def _maybe_start_eip(self, autostart=False):
        """
        Start the EIP bootstrapping sequence if the client is configured to
        do so.

        :param autostart: we are autostarting EIP when this is True
        :type autostart: bool
        """
        # XXX should move to EIP conductor.

        # during autostart we assume that the provider provides EIP
        if autostart:
            should_start = EIP_SERVICE in self._enabled_services
        else:
            should_start = self._provides_eip_and_enabled()

        missing_helpers = self._eip_status.missing_helpers
        already_started = self._already_started_eip
        can_start = (should_start and
                     not already_started and
                     not missing_helpers)

        if can_start:
            if self._eip_status.is_cold_start:
                self._backend.tear_fw_down()
            # XXX this should be handled by the state machine.
            self._enable_eip_start_action()
            self._eip_status.set_eip_status(
                self.tr("Starting..."))
            self._eip_status.show_eip_cancel_button()

            # We were disabling the button, but now that we have
            # a cancel button we just hide it. It will be visible
            # when the connection is completed successfully.
            self._eip_status.eip_button.hide()
            self._eip_status.eip_button.setEnabled(False)

            domain = self._providers.get_selected_provider()
            self._backend.eip_setup(provider=domain)

            self._already_started_eip = True
            # we want to start soledad anyway after a certain timeout if eip
            # fails to come up
            QtDelayedCall(self.EIP_START_TIMEOUT,
                          self._maybe_run_soledad_setup_checks)
        else:
            if not self._already_started_eip:
                if EIP_SERVICE in self._enabled_services:
                    if missing_helpers:
                        msg = self.tr("Disabled: missing helper files")
                    else:
                        msg = self.tr("Not supported"),
                    self._eip_status.set_eip_status(msg, error=True)
                else:
                    msg = self.tr("Disabled")
                    self._eip_status.disable_eip_start()
                    self._eip_status.set_eip_status(msg)

            # eip will not start, so we start soledad anyway
            # XXX This is the entry point for soledad startup.
            self._maybe_run_soledad_setup_checks()

    def _finish_eip_bootstrap(self, data):
        """
        TRIGGERS:
            self._backend.signaler.eip_client_certificate_ready

        Start the VPN thread if the eip configuration is properly
        loaded.
        """
        passed = data[PASSED_KEY]

        if not passed:
            error_msg = self.tr("There was a problem with the provider")
            self._eip_status.set_eip_status(error_msg, error=True)
            logger.error(data[ERROR_KEY])
            self._already_started_eip = False
            return

        # DO START EIP Connection!
        self._eip_conductor.do_connect()

    def _eip_intermediate_stage(self, data):
        # TODO missing param documentation
        """
        TRIGGERS:
            self._backend.signaler.eip_config_ready

        If there was a problem, displays it, otherwise it does nothing.
        This is used for intermediate bootstrapping stages, in case
        they fail.
        """
        passed = data[PASSED_KEY]
        if not passed:
            self._eip_status.set_eip_status(
                self.tr("Unable to connect: Problem with provider"),
                error=True)
            logger.error(data[ERROR_KEY])
            self._already_started_eip = False
            self._eip_status.aborted()

    # end of EIP methods ---------------------------------------------

    #
    # window handling methods
    #

    def _on_raise_window_event(self, event):
        """
        Callback for the raise window event

        :param event: The event that triggered the callback.
        :type event: str
        """
        if IS_WIN:
            locks.raise_window_ack()
        self.raise_window.emit()

    def _do_raise_mainwindow(self):
        """
        TRIGGERS:
            self._on_raise_window_event

        Triggered when we receive a RAISE_WINDOW event.
        """
        TOPFLAG = QtCore.Qt.WindowStaysOnTopHint
        self.setWindowFlags(self.windowFlags() | TOPFLAG)
        self.show()
        self.setWindowFlags(self.windowFlags() & ~TOPFLAG)
        self.show()
        if IS_MAC:
            self.raise_()

    #
    # cleanup and quit methods
    #

    def _stop_services(self):
        """
        Stop services and cancel ongoing actions (if any).
        """
        logger.debug('Stopping services...')

        self._cancel_ongoing_defers()
        self._services_being_stopped = set(('imap', 'eip'))

        imap_stopped = lambda: self._remove_service('imap')
        self._leap_signaler.imap_stopped.connect(imap_stopped)

        # XXX change name, already used in conductor.
        eip_stopped = lambda: self._remove_service('eip')
        self._leap_signaler.eip_stopped.connect(eip_stopped)

        logger.debug('Stopping mail services')
        self._mail_conductor.stop_mail_services()

        logger.debug("Doing logout")
        self._login_widget.do_logout()

        logger.debug('Terminating vpn')
        self._backend.eip_stop(shutdown=True)

    def quit(self, disable_autostart=True):
        """
        Start the quit sequence and wait for services to finish.
        Cleanup and close the main window before quitting.

        :param disable_autostart: whether we should disable the autostart
                                  feature or not
        :type disable_autostart: bool
        """
        if self._quitting:
            return

        if disable_autostart:
            autostart.set_autostart(False)

        self._quitting = True
        self._close_to_tray = False
        logger.debug('Quitting...')

        # first thing to do quitting, hide the mainwindow and show tooltip.
        self.hide()
        if not self._system_quit and self._systray is not None:
            self._systray.showMessage(
                self.tr('Quitting...'),
                self.tr('Bitmask is quitting, please wait.'))

        # explicitly process events to display tooltip immediately
        QtCore.QCoreApplication.processEvents(0, 10)

        # Close other windows if any.
        if self._wizard:
            self._wizard.close()

        # Set this in case that the app is hidden
        QtGui.QApplication.setQuitOnLastWindowClosed(True)

        if not self._backend.online:
            self.final_quit()
            return

        # call final quit when all the services are stopped
        self.all_services_stopped.connect(self.final_quit)

        self._stop_services()

        # we wait and call manually since during the system's logout the
        # backend process can be killed and we won't get a response.
        # XXX: also, for some reason the services stop timeout does not work.
        if self._system_quit:
            time.sleep(0.5)
            self.final_quit()

        # or if we reach the timeout
        QtDelayedCall(self.SERVICES_STOP_TIMEOUT, self.final_quit)

    def _backend_kill(self):
        """
        Send a kill signal to the backend process.
        This is called if the backend does not respond to requests.
        """
        if self._backend_pid is not None:
            logger.debug("Killing backend")
            psutil.Process(self._backend_pid).kill()

    def _remove_service(self, service):
        """
        Remove the given service from the waiting list and check if we have
        running services that we need to wait until we quit.
        Emit self.all_services_stopped signal if we don't need to keep waiting.

        :param service: the service that we want to remove
        :type service: str
        """
        logger.debug("Removing service: {0}".format(service))
        self._services_being_stopped.discard(service)

        if not self._services_being_stopped:
            logger.debug("All services stopped.")
            self.all_services_stopped.emit()

    def final_quit(self):
        """
        Final steps to quit the app, starting from here we don't care about
        running services or user interaction, just quitting.
        """
        # We can reach here because all the services are stopped or because a
        # timeout was triggered. Since we want to run this only once, we exit
        # if this is called twice.
        if self._finally_quitting:
            return

        logger.debug('Final quit...')
        self._finally_quitting = True

        if self._backend.online:
            logger.debug('Closing soledad...')
            self._backend.soledad_close()

        self._leap_signaler.stop()

        LOG_CONTROLLER.stop_logbook_subscriber()
        self._backend.stop()
        time.sleep(0.05)  # give the thread a little time to finish.

        if self._system_quit or not self._backend.online:
            logger.debug("Killing the backend")
            self._backend_kill()

        # Remove lockfiles on a clean shutdown.
        logger.debug('Cleaning pidfiles')
        locks.release_lock()

        self.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# setup.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Setup file for bitmask.
"""
from __future__ import print_function

import hashlib
import sys
import os
import re
import sys

if not sys.version_info[0] == 2:
    print("[ERROR] Sorry, Python 3 is not supported (yet). "
          "Try running with python2: python2 setup.py ...")
    exit()

try:
    from setuptools import setup, find_packages
except ImportError:
    from pkg import distribute_setup
    distribute_setup.use_setuptools()
    from setuptools import setup, find_packages

from pkg import utils

import versioneer
versioneer.versionfile_source = 'src/leap/bitmask/_version.py'
versioneer.versionfile_build = 'leap/bitmask/_version.py'
versioneer.tag_prefix = ''  # tags are like 1.2.0
versioneer.parentdir_prefix = 'leap.bitmask-'


# The following import avoids the premature unloading of the `util` submodule
# when running tests, which would cause an error when nose finishes tests and
# calls the exit function of the multiprocessing module.
from multiprocessing import util
assert(util)

setup_root = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(setup_root, "src"))

trove_classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: X11 Applications :: Qt",
    "Intended Audience :: End Users/Desktop",
    ("License :: OSI Approved :: GNU General "
     "Public License v3 or later (GPLv3+)"),
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Topic :: Security",
    'Topic :: Security :: Cryptography',
    "Topic :: Communications",
    'Topic :: Communications :: Email',
    'Topic :: Communications :: Email :: Post-Office :: IMAP',
    'Topic :: Internet',
    "Topic :: Utilities",
]

DOWNLOAD_BASE = ('https://github.com/leapcode/bitmask_client/'
                 'archive/%s.tar.gz')
_versions = versioneer.get_versions()
VERSION = _versions['version']
VERSION_FULL = _versions['full']
DOWNLOAD_URL = ""

# get the short version for the download url
_version_short = re.findall('\d+\.\d+\.\d+', VERSION)
if len(_version_short) > 0:
    VERSION_SHORT = _version_short[0]
    DOWNLOAD_URL = DOWNLOAD_BASE % VERSION_SHORT

cmdclass = versioneer.get_cmdclass()


from setuptools import Command


class freeze_debianver(Command):

    """
    Freezes the version in a debian branch.
    To be used after merging the development branch onto the debian one.
    """
    user_options = []
    template = r"""
# This file was generated by the `freeze_debianver` command in setup.py
# Using 'versioneer.py' (0.7+) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '{version}'
version_full = '{version_full}'
"""
    templatefun = r"""

def get_versions(default={}, verbose=False):
        return {'version': version_version, 'full': version_full}
"""

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        proceed = str(raw_input(
            "This will overwrite the file _version.py. Continue? [y/N] "))
        if proceed != "y":
            print("He. You scared. Aborting.")
            return
        subst_template = self.template.format(
            version=VERSION_SHORT,
            version_full=VERSION_FULL) + self.templatefun
        with open(versioneer.versionfile_source, 'w') as f:
            f.write(subst_template)


def freeze_pkg_ver(path, version_short, version_full):
    """
    Freeze the _version in other modules, used during the gathering of
    all the leap modules in the sumo tarball.
    """
    subst_template = freeze_debianver.template.format(
        version=version_short,
        version_full=version_full) + freeze_debianver.templatefun
    with open(path, 'w') as f:
        f.write(subst_template)


if sys.argv[:1] == '--sumo':
    IS_SUMO = True
else:
    IS_SUMO = False

cmdclass["freeze_debianver"] = freeze_debianver
parsed_reqs = utils.parse_requirements()

if utils.is_develop_mode() or IS_SUMO:
    print("")
    print ("[WARNING] Skipping leap-specific dependencies "
           "because development mode is detected.")
    print ("[WARNING] You can install "
           "the latest published versions with "
           "'pip install -r pkg/requirements-leap.pip'")
    print ("[WARNING] Or you can instead do 'python setup.py develop' "
           "from the parent folder of each one of them.")
    print("")
else:
    parsed_reqs += utils.parse_requirements(
        reqfiles=["pkg/requirements-leap.pip"])


leap_launcher = 'bitmask=leap.bitmask.app:start_app'

from setuptools.command.develop import develop as _develop


def copy_reqs(path, withsrc=False):
    # add a copy of the processed requirements to the package
    _reqpath = ('leap', 'bitmask', 'util', 'reqs.txt')
    if withsrc:
        reqsfile = os.path.join(path, 'src', *_reqpath)
    else:
        reqsfile = os.path.join(path, *_reqpath)
    print("UPDATING %s" % reqsfile)
    if os.path.isfile(reqsfile):
        os.unlink(reqsfile)
    with open(reqsfile, "w") as f:
        f.write('\n'.join(parsed_reqs))


class cmd_develop(_develop):

    def run(self):
        # versioneer:
        versions = versioneer.get_versions(verbose=True)
        self._versioneer_generated_versions = versions
        # unless we update this, the command will keep using the old version
        self.distribution.metadata.version = versions["version"]

        _develop.run(self)
        copy_reqs(self.egg_path)

cmdclass["develop"] = cmd_develop


class cmd_binary_hash(Command):

    """
    Update the _binaries.py file with hashes for the different helpers.
    This is used from within the bundle.
    """

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self, *args):

        OPENVPN_BIN = os.environ.get('OPENVPN_BIN', None)
        BITMASK_ROOT = os.environ.get('BITMASK_ROOT', None)

        def exit():
            print("Please set environment variables "
                  "OPENVPN_BIN and BITMASK_ROOT pointing to the right path "
                  "to use this command")
            sys.exit(1)

        bin_paths = OPENVPN_BIN, BITMASK_ROOT
        if not all(bin_paths):
            exit()

        if not all(map(os.path.isfile, bin_paths)):
            exit()

        openvpn_bin_hash, bitmask_root_hash = map(
            lambda path: hashlib.sha256(open(path).read()).hexdigest(),
            bin_paths)

        template = r"""
# Hashes for binaries used in Bitmask Bundle.
# This file has been automatically generated by `setup.py hash_binaries`
# DO NOT modify it manually.

OPENVPN_BIN = "{openvpn}"
BITMASK_ROOT = "{bitmask}"
"""
        subst_template = template.format(
            openvpn=openvpn_bin_hash,
            bitmask=bitmask_root_hash)

        bin_hash_path = os.path.join('src', 'leap', 'bitmask', '_binaries.py')
        with open(bin_hash_path, 'w') as f:
            f.write(subst_template)
        print("Binaries hash file %s has been updated!" % (bin_hash_path,))


cmdclass["hash_binaries"] = cmd_binary_hash


# next two classes need to augment the versioneer modified ones

versioneer_build = cmdclass['build']
versioneer_sdist = cmdclass['sdist']


class cmd_build(versioneer_build):

    def run(self):
        versioneer_build.run(self)
        copy_reqs(self.build_lib)


class cmd_sdist(versioneer_sdist):

    user_options = versioneer_sdist.user_options + \
        [('sumo', 's',
          "create a 'sumo' sdist which includes the contents of all "
          "the leap.* packages")
         ]
    boolean_options = ['sumo']
    leap_sumo_packages = ['soledad.common', 'soledad.client',
                          'keymanager', 'mail', 'common']

    def initialize_options(self):
        versioneer_sdist.initialize_options(self)
        self.sumo = False

    def run(self):
        return versioneer_sdist.run(self)

    def make_release_tree(self, base_dir, files):
        versioneer_sdist.make_release_tree(self, base_dir, files)
        # We need to copy the requirements to the specified path
        # so that the client has a copy to do the startup checks.
        copy_reqs(base_dir, withsrc=True)
        try:
            with open(os.path.join(base_dir,
                                   'src', 'leap', '__init__.py'),
                      'w') as nuke_top_init:
                nuke_top_init.write('')
        except Exception:
            pass
        try:
            with open(os.path.join(base_dir,
                                   'src', 'leap', 'soledad', '__init__.py'),
                      'w') as nuke_soledad_ns:
                nuke_soledad_ns.write('')
        except Exception:
            pass

    def make_distribution(self):
        # add our extra files to the list just before building the
        # tarball/zipfile. We override make_distribution() instead of run()
        # because setuptools.command.sdist.run() does not lend itself to
        # easy/robust subclassing (the code we need to add goes right smack
        # in the middle of a 12-line method). If this were the distutils
        # version, we'd override get_file_list().

        if self.sumo:
            # If '--sumo' was specified, include all the leap.* in the sdist.
            vdict = _get_leap_versions()
            vdict['soledad.common'] = vdict['soledad']
            vdict['soledad.client'] = vdict['soledad']
            import importlib
            for module in self.leap_sumo_packages:
                full_module = "leap." + module
                importlib.import_module(full_module)
                src_path = "src/leap/" + _fix_namespace(module)
                imported_module = importlib.sys.modules[full_module]
                copy_recursively(
                    imported_module.__path__[0] + "/",
                    src_path)
                all_module_files = list_recursively(src_path)
                self.filelist.extend(all_module_files)
                module_ver = vdict[module]
                freeze_pkg_ver(
                    src_path + "/_version.py",
                    module_ver, "%s-sumo" % module_ver)
            freeze_pkg_ver(
                "src/leap/bitmask/_version.py",
                VERSION, "%s-sumo" % VERSION)

            # In addition, we want the tarball/zipfile to have -SUMO in the
            # name, and the unpacked directory to have -SUMO too. The easiest
            # way to do this is to patch self.distribution and override the
            # get_fullname() method. (an alternative is to modify
            # self.distribution.metadata.version, but that also affects the
            # contents of PKG-INFO).
            fullname = self.distribution.get_fullname()

            def get_fullname():
                return fullname + "-SUMO"
            self.distribution.get_fullname = get_fullname

        try:
            old_mask = os.umask(int("022", 8))
            return versioneer_sdist.make_distribution(self)
        finally:
            os.umask(old_mask)
            for module in self.leap_sumo_packages:
                # check, just in case...
                if module and module != "bitmask":
                    try:
                        shutil.rmtree("src/leap/" + _fix_namespace(module))
                    except Exception:
                        pass


import shutil
import glob


def _get_leap_versions():
    versions = {}
    with open("pkg/leap_versions.txt") as vf:
        lines = vf.readlines()
    for line in lines:
        pkg, ver = line.split('\t')
        versions[pkg.strip().replace('leap_', '')] = ver.strip()
    return versions


def _fix_namespace(path):
    if path in ('soledad.common', 'soledad.client'):
        return path.replace('.', '/')
    return path


_ignore_files = ('*.pyc', '_trial*', '*.swp', '.*', 'cert', 'test*')
_ignore_dirs = ('tests', '_trial*', 'test*')
_ignore_paths = _ignore_files + _ignore_dirs

is_excluded_path = lambda path: any(
    map(lambda pattern: glob.fnmatch.fnmatch(path, pattern),
        _ignore_paths))


def _should_exclude(path):
    folder, f = os.path.split(path)
    if is_excluded_path(f):
        return True
    upper, leaf = os.path.split(folder)
    if is_excluded_path(leaf):
        return True
    return False


def list_recursively(root_dir):
    file_list = []
    for root, sub_dirs, files in os.walk(root_dir):
        for f in files:
            is_excluded = _should_exclude(f)
            if not is_excluded:
                file_list.append(os.path.join(root, f))
    return file_list


def _mkdir_recursively(path):
    sub_path = os.path.dirname(path)
    if not os.path.exists(sub_path):
        _mkdir_recursively(sub_path)
    if not os.path.exists(path):
        os.mkdir(path)


def copy_recursively(source_folder, destination_folder):

    if not os.path.exists(destination_folder):
        _mkdir_recursively(destination_folder)

    for root, dirs, files in os.walk(source_folder):
        if _should_exclude(root):
            continue
        for item in files:
            if _should_exclude(item):
                continue
            src_path = os.path.join(root, item)
            dst_path = os.path.join(
                destination_folder, src_path.replace(source_folder, ""))
            if _should_exclude(dst_path):
                continue
            if os.path.exists(dst_path):
                if os.stat(src_path).st_mtime > os.stat(dst_path).st_mtime:
                    shutil.copy2(src_path, dst_path)
            else:
                shutil.copy2(src_path, dst_path)
        for item in dirs:
            if _should_exclude(item):
                continue
            src_path = os.path.join(root, item)
            dst_path = os.path.join(
                destination_folder, src_path.replace(source_folder, ""))
            if _should_exclude(dst_path):
                continue
            if not os.path.exists(dst_path):
                os.mkdir(dst_path)

cmdclass["build"] = cmd_build
cmdclass["sdist"] = cmd_sdist

import platform
_system = platform.system()
IS_LINUX = _system == "Linux"
IS_MAC = _system == "Darwin"

data_files = []


if IS_LINUX:
    # XXX use check_for_permissions to install data
    # globally. Or make specific install command. See #3805
    isset = lambda var: os.environ.get(var, None)
    if isset('VIRTUAL_ENV') or isset('LEAP_SKIP_INIT'):
        data_files = None
    else:
        data_files = [
            ("share/polkit-1/actions",
                ["pkg/linux/polkit/se.leap.bitmask.policy"]),
            ("/usr/sbin",
                ["pkg/linux/bitmask-root"]),
        ]

extra_options = {}

setup(
    name="leap.bitmask",
    package_dir={"": "src"},
    version=VERSION,
    cmdclass=cmdclass,
    description=("The Internet Encryption Toolkit: "
                 "Encrypted Internet Proxy and Encrypted Mail."),
    long_description=open('README.rst').read() + '\n\n\n' +
    open('CHANGELOG.rst').read(),
    classifiers=trove_classifiers,
    install_requires=parsed_reqs,
    test_suite='nose.collector',
    tests_require=utils.parse_requirements(
        reqfiles=['pkg/requirements-testing.pip']),
    keywords=('Bitmask, LEAP, client, qt, encryption, '
              'proxy, openvpn, imap, smtp, gnupg'),
    author='The LEAP Encryption Access Project',
    author_email='info@leap.se',
    maintainer='Kali Kaneko',
    maintainer_email='kali@leap.se',
    url='https://bitmask.net',
    download_url=DOWNLOAD_URL,
    license='GPL-3+',
    packages=find_packages(
        'src',
        exclude=['ez_setup', 'setup', 'examples', 'tests']),
    namespace_packages=["leap"],
    package_data={'': ['util/*.txt', '*.pem']},
    include_package_data=True,
    # not being used? -- setuptools does not like it.
    # looks like debhelper is honoring it...
    data_files=data_files,
    zip_safe=False,
    platforms="all",
    entry_points={
        'console_scripts': [leap_launcher]
    },
    **extra_options
)



# -*- coding: utf-8 -*-
# passwordwindow.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Change password dialog window
"""

from PySide import QtGui

from leap.bitmask.logs.utils import get_logger
from leap.bitmask.util.credentials import password_checks
from leap.bitmask.gui.ui_password_change import Ui_PasswordChange
from leap.bitmask.gui.flashable import Flashable

logger = get_logger()


class PasswordWindow(QtGui.QDialog, Flashable):

    _current_window = None  # currently visible password window

    def __init__(self, parent, account, app):
        """
        :param parent: parent object of the PreferencesWindow.
        :parent type: QWidget

        :param account: the user set in the login widget
        :type account: Account

        :param app: App instance
        :type app: App
        """
        QtGui.QDialog.__init__(self, parent)

        self.account = account
        self.app = app
        self._backend_connect()

        self.ui = Ui_PasswordChange()
        self.ui.setupUi(self)

        self.hide_flash()
        self.ui.ok_button.clicked.connect(self._change_password)
        self.ui.cancel_button.clicked.connect(self.close)
        self.ui.username_lineedit.setText(account.address)

        if PasswordWindow._current_window is not None:
            PasswordWindow._current_window.close()
        PasswordWindow._current_window = self

        self._disabled = False  # if set to True, never again enable widgets.

        if account.username is None:
            # should not ever happen, but just in case
            self._disabled = True
            self._enable_password_widgets(False)
            self.ui.cancel_button.setEnabled(True)
            self.flash_error(self.tr("Please log in to change your password."))

        if self.is_soledad_needed() and not self.app.soledad_started:
            self._enable_password_widgets(False)
            self.ui.cancel_button.setEnabled(True)

            self.flash_message(
                self.tr("Please wait for data storage to be ready."))

    def is_soledad_needed(self):
        """
        Returns true if the current account needs to change the soledad
        password as well as the SRP password.
        """
        return self.account.is_email_enabled()

    #
    # MANAGE WIDGETS
    #

    def _enable_password_widgets(self, enabled):
        """
        Enables or disables the widgets in the password change group box.

        :param enabled: True if the widgets should be enabled.
                        False if widgets should be disabled and
                        display the status label that shows that is
                        changing the password.
        :type enabled: bool
        """
        if self._disabled:
            return

        if enabled:
            self.hide_flash()
        else:
            self.flash_message(self.tr("Changing password..."))

        self.ui.current_password_lineedit.setEnabled(enabled)
        self.ui.new_password_lineedit.setEnabled(enabled)
        self.ui.new_password_confirmation_lineedit.setEnabled(enabled)
        self.ui.ok_button.setEnabled(enabled)
        self.ui.cancel_button.setEnabled(enabled)

    def _change_password_success(self):
        """
        Callback used to display a successfully changed password.
        """
        logger.debug("Password changed successfully.")
        self._clear_password_inputs()
        self._enable_password_widgets(True)
        self.flash_success(self.tr("Password changed successfully."))

    def _clear_password_inputs(self):
        """
        Clear the contents of the inputs.
        """
        self.ui.current_password_lineedit.setText("")
        self.ui.new_password_lineedit.setText("")
        self.ui.new_password_confirmation_lineedit.setText("")

    #
    # SLOTS
    #

    def _backend_connect(self):
        """
        Helper to connect to backend signals
        """
        sig = self.app.signaler
        sig.srp_password_change_ok.connect(self._srp_change_password_ok)
        sig.srp_password_change_error.connect(self._srp_password_change_error)
        sig.srp_password_change_badpw.connect(self._srp_password_change_badpw)
        sig.soledad_password_change_ok.connect(
            self._soledad_change_password_ok)
        sig.soledad_password_change_error.connect(
            self._soledad_change_password_problem)

        sig.soledad_bootstrap_finished.connect(self._on_soledad_ready)

    def _change_password(self):
        """
        TRIGGERS:
            self.ui.buttonBox.accepted

        Changes the user's password if the inputboxes are correctly filled.
        """
        current_password = self.ui.current_password_lineedit.text()
        new_password = self.ui.new_password_lineedit.text()
        new_password2 = self.ui.new_password_confirmation_lineedit.text()

        self._enable_password_widgets(True)

        if len(current_password) == 0:
            self.flash_error(self.tr("Password is empty."))
            self.ui.current_password_lineedit.setFocus()
            return

        ok, msg, field = password_checks(self.account.username, new_password,
                                         new_password2)
        if not ok:
            self.flash_error(msg)
            if field == 'new_password':
                self.ui.new_password_lineedit.setFocus()
            elif field == 'new_password_confirmation':
                self.ui.new_password_confirmation_lineedit.setFocus()
            return

        self._enable_password_widgets(False)
        self.app.backend.user_change_password(
            current_password=current_password,
            new_password=new_password)

    def closeEvent(self, event=None):
        """
        TRIGGERS:
            cancel_button (indirectly via self.close())
            or when window is closed

        Close this dialog & delete ourselves to clean up signals.
        """
        PasswordWindow._current_window = None
        self.deleteLater()

    def _srp_change_password_ok(self):
        """
        TRIGGERS:
            self._backend.signaler.srp_password_change_ok

        Callback used to display a successfully changed password.
        """
        new_password = self.ui.new_password_lineedit.text()
        logger.debug("SRP password changed successfully.")

        # FIXME ---- both changes need to be made atomically!
        # if there is some problem changing password in soledad (for instance,
        # it checks for length), any exception raised will be lost and we will
        # have an inconsistent state between soledad and srp passwords.
        # We need to implement rollaback.

        if self.is_soledad_needed():
            self.app.backend.soledad_change_password(new_password=new_password)
        else:
            self._change_password_success()

    def _srp_password_change_error(self):
        """
        TRIGGERS:
            self._backend.signaler.srp_password_change_error

        Unknown problem changing password
        """
        msg = self.tr("There was a problem changing the password.")
        logger.error(msg)
        self._enable_password_widgets(True)
        self.flash_error(msg)

    def _srp_password_change_badpw(self):
        """
        TRIGGERS:
            self._backend.signaler.srp_password_change_badpw

        The password the user entered was wrong.
        """
        msg = self.tr("You did not enter a correct current password.")
        logger.error(msg)
        self._enable_password_widgets(True)
        self.flash_error(msg)
        self.ui.current_password_lineedit.setFocus()

    def _soledad_change_password_ok(self):
        """
        TRIGGERS:
            Signaler.soledad_password_change_ok

        Soledad password change went OK.
        """
        logger.debug("Soledad password changed successfully.")
        self._change_password_success()

    def _soledad_change_password_problem(self, msg):
        """
        TRIGGERS:
            Signaler.soledad_password_change_error

        Callback used to display an error on changing password.

        :param msg: the message to show to the user.
        :type msg: unicode
        """
        logger.error("Error changing soledad password: %s" % (msg,))
        self._enable_password_widgets(True)
        self.flash_error(msg)

    def _on_soledad_ready(self):
        """
        TRIGGERS:
            Signaler.soledad_bootstrap_finished
        """
        self._enable_password_widgets(True)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'src/leap/bitmask/gui/ui/preferences.ui'
#
# Created: Mon Dec 29 09:53:14 2014
#      by: pyside-uic 0.2.15 running on PySide 1.2.2
#
# WARNING! All changes made in this file will be lost!

from PySide import QtCore, QtGui

class Ui_Preferences(object):
    def setupUi(self, Preferences):
        Preferences.setObjectName("Preferences")
        Preferences.resize(520, 439)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/images/mask-icon.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        Preferences.setWindowIcon(icon)
        self.verticalLayout = QtGui.QVBoxLayout(Preferences)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.account_label = QtGui.QLabel(Preferences)
        font = QtGui.QFont()
        font.setWeight(75)
        font.setBold(True)
        self.account_label.setFont(font)
        self.account_label.setObjectName("account_label")
        self.verticalLayout.addWidget(self.account_label)
        self.line = QtGui.QFrame(Preferences)
        self.line.setFrameShape(QtGui.QFrame.HLine)
        self.line.setFrameShadow(QtGui.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.horizontal_layout = QtGui.QHBoxLayout()
        self.horizontal_layout.setSpacing(12)
        self.horizontal_layout.setObjectName("horizontal_layout")
        self.nav_widget = QtGui.QListWidget(Preferences)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Maximum, QtGui.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.nav_widget.sizePolicy().hasHeightForWidth())
        self.nav_widget.setSizePolicy(sizePolicy)
        self.nav_widget.setMaximumSize(QtCore.QSize(120, 16777215))
        self.nav_widget.setIconSize(QtCore.QSize(32, 32))
        self.nav_widget.setMovement(QtGui.QListView.Static)
        self.nav_widget.setSpacing(10)
        self.nav_widget.setViewMode(QtGui.QListView.IconMode)
        self.nav_widget.setUniformItemSizes(True)
        self.nav_widget.setObjectName("nav_widget")
        self.horizontal_layout.addWidget(self.nav_widget)
        self.pages_widget = QtGui.QStackedWidget(Preferences)
        self.pages_widget.setObjectName("pages_widget")
        self.horizontal_layout.addWidget(self.pages_widget)
        self.verticalLayout.addLayout(self.horizontal_layout)
        self.button_layout = QtGui.QHBoxLayout()
        self.button_layout.setObjectName("button_layout")
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.button_layout.addItem(spacerItem)
        self.close_button = QtGui.QPushButton(Preferences)
        self.close_button.setObjectName("close_button")
        self.button_layout.addWidget(self.close_button)
        self.verticalLayout.addLayout(self.button_layout)
        self.verticalLayout.setStretch(2, 1)

        self.retranslateUi(Preferences)
        QtCore.QMetaObject.connectSlotsByName(Preferences)

    def retranslateUi(self, Preferences):
        Preferences.setWindowTitle(QtGui.QApplication.translate("Preferences", "Preferences", None, QtGui.QApplication.UnicodeUTF8))
        self.account_label.setText(QtGui.QApplication.translate("Preferences", "user@example.org", None, QtGui.QApplication.UnicodeUTF8))
        self.close_button.setText(QtGui.QApplication.translate("Preferences", "Close", None, QtGui.QApplication.UnicodeUTF8))

import icons_rc
